# In[Imports]

import numpy as np
import xarray as xr
from os.path import expanduser

# Functions to make the Equation of State
from neutralocean.eos import make_eos, make_eos_s_t

# Functions to compute various approximately neutral surfaces
from neutralocean.surface import potential_surf, anomaly_surf, omega_surf

from neutralocean.grid.xgcm import build_grid, edgedata_to_maps
from neutralocean.ntp import ntp_epsilon_errors

# In[Load data]

print(
    "To get started, download the ECCOv4r4 grid information at\n"
    "  https://archive.podaac.earthdata.nasa.gov/podaac-ops-cumulus-protected/ECCO_L4_GEOMETRY_LLC0090GRID_V4R4/GRID_GEOMETRY_ECCO_V4r4_native_llc0090.nc"
    "\nand one day of Salinity and Temperature 3D data at\n"
    "  https://archive.podaac.earthdata.nasa.gov/podaac-ops-cumulus-protected/ECCO_L4_TEMP_SALINITY_LLC0090GRID_DAILY_V4R4/OCEAN_TEMPERATURE_SALINITY_day_mean_2002-12-23_ECCO_V4r4_native_llc0090.nc"
    "\nYou will have to create an Earthdata Login first.  "
    "Then edit the below variable, `folder_ecc4`, to point to the directory"
    " where you saved these two files."
)

folder_ecco4 = expanduser("~/work/data/ECCOv4r4/")  # << EDIT AS NEEDED >>
file_grid = folder_ecco4 + "GRID_GEOMETRY_ECCO_V4r4_native_llc0090.nc"
date = "2002-12-23"
file_ST = (
    folder_ecco4
    + "OCEAN_TEMPERATURE_SALINITY_day_mean_"
    + date
    + "_ECCO_V4r4_native_llc0090.nc"
)

# Load horizontal grid information
ds = xr.open_dataset(file_grid)
XC, YC, dxC, dyC, dyG, dxG = (
    ds[x].load() for x in ("XC", "YC", "dxC", "dyC", "dyG", "dxG")
)
ds.close()

# Load hydrographic data
ds = xr.open_dataset(file_ST)
S, T = (ds[x].squeeze().load() for x in ("SALT", "THETA"))

# Create depth xarray.
Z = -ds.Z.load()  # Make Z > 0 and increasing down
Z.attrs.update({"positive": "down"})  # Update attrs to match.

n = len(ds.i)  # size of each horizontal dimension in each square tile
ds.close()

# Get order of non-vertical dimensions
dims = tuple(x for x in S.dims if x != "k")

# Reorder dimensions to ensure individual water columns are float64 and contiguous in memory
S, T = (x.transpose(*dims, "k").astype(np.float64, order="C") for x in (S, T))

# define the connectivity between faces for the ECCOv4 LLC grid:
# fmt: off
face_connections = {'tile': {
    0: {'X':  ((12, 'Y', False), (3, 'X', False)),
        'Y':  (None,             (1, 'Y', False))},
    1: {'X':  ((11, 'Y', False), (4, 'X', False)),
        'Y':  ((0, 'Y', False),  (2, 'Y', False))},
    2: {'X':  ((10, 'Y', False), (5, 'X', False)),
        'Y':  ((1, 'Y', False),  (6, 'X', False))},
    3: {'X':  ((0, 'X', False),  (9, 'Y', False)),
        'Y':  (None,             (4, 'Y', False))},
    4: {'X':  ((1, 'X', False),  (8, 'Y', False)),
        'Y':  ((3, 'Y', False),  (5, 'Y', False))},
    5: {'X':  ((2, 'X', False),  (7, 'Y', False)),
        'Y':  ((4, 'Y', False),  (6, 'Y', False))},
    6: {'X':  ((2, 'Y', False),  (7, 'X', False)),
        'Y':  ((5, 'Y', False),  (10, 'X', False))},
    7: {'X':  ((6, 'X', False),  (8, 'X', False)),
        'Y':  ((5, 'X', False),  (10, 'Y', False))},
    8: {'X':  ((7, 'X', False),  (9, 'X', False)),
        'Y':  ((4, 'X', False),  (11, 'Y', False))},
    9: {'X':  ((8, 'X', False),  None),
        'Y':  ((3, 'X', False),  (12, 'Y', False))},
    10:{'X': ((6, 'Y', False),  (11, 'X', False)),
        'Y': ((7, 'Y', False),  (2, 'X', False))},
    11:{'X': ((10, 'X', False), (12, 'X', False)),
        'Y': ((8, 'Y', False),  (1, 'X', False))},
    12:{'X': ((11, 'X', False), None),
        'Y': ((9, 'Y', False),  (0, 'X', False))}}}
# fmt: on

# Number of Faces -- len(faces_connections only value)
nf = len(next(iter(face_connections.values())))

# Build list of adjacent water columns and distances between those water column pairs
xsh = ysh = "left"
grid = build_grid(n, face_connections, dims, xsh, ysh, dxC, dyC, dyG, dxG)


# Make Boussinesq version of the Jackett and McDougall (1995) equation of state
#  and its partial derivatives.
# TODO: is this what ECCOv4r4 used?
grav, rho_c = 9.81, 1027.5
eos = make_eos("jmd95", grav, rho_c)
eos_s_t = make_eos_s_t("jmd95", grav, rho_c)

# Select pinning cast, picking the cast closest to (x0,y0)
x0, y0 = (-172, -4)  # longitude, latitude -- Pacific equatorial ocean
pin_cast = np.unravel_index(
    ((XC.values - x0) ** 2 + (YC.values - y0) ** 2).argmin(), XC.shape
)
z0 = 1500.0  # pinning depth

# In[Build approximately neutral surfaces]

# Build potential density surface, with given reference pressure (actually depth,
# for Boussinesq) and given isovalue.  No diagnostics requested, so info about
# the grid is not needed (no `edges` and `geoemtry` provided), and also eos_s_t
# is not needed.
s, t, z, _ = potential_surf(
    S, T, Z, eos=eos, vert_dim="k", ref=0.0, isoval=1027.5, diags=False
)
z_sigma = z

# Build in-situ density anomaly surface with given reference salinity and
# potential temperature values and an isovalue of 0, which means the surface
# will intersect any point where the local (S,T) equals the reference values.
# Also return diagnostics.
s0, t0 = 34.5, 4.0
s, t, z, d = anomaly_surf(
    S,
    T,
    Z,
    grid=grid,
    eos=(eos, eos_s_t),
    vert_dim="k",
    ref=(s0, t0),
    isoval=0.0,
)

# Build an omega surface that is initialized from the above potential density
# surface and is pinned at the cast `pin_cast` (i.e. the omega surface will have
# the same depth as the initializing potential density surface at this cast).
s, t, z, d = omega_surf(
    S,
    T,
    Z,
    grid=grid,
    vert_dim="k",
    p_init=z_sigma,
    pin_cast=pin_cast,
    eos=(eos, eos_s_t),
    interp="pchip",
    ITER_MAX=10,
    ITER_START_WETTING=1,
)
z_omega = z

# In[Calculate neutrality error]

# Calculate ϵ neutrality errors on the latest surface, between all pairs of adjacent water columns
e = ntp_epsilon_errors(s, t, z, grid, eos_s_t)

# Convert the 1D array of ϵ values into two maps of ϵ neutrality errors, one
# for the errors in each of the two lateral ('i' and 'j') dimensions.  These
# neutrality errors can then be mapped or further analyzed.
ei, ej = edgedata_to_maps(e, n, face_connections, dims, xsh, ysh)

# In[Map depth difference between omega and potential]

# I've installed ecco_v4_py by cloning their git repository to ~/work/python/ECCOv4-py
# Follow https://ecco-v4-python-tutorial.readthedocs.io/Installing_Python_and_Python_Packages.html#option-1-clone-into-the-repository-using-git-recommended
# and edit path below as needed.
import sys
import matplotlib.pyplot as plt

sys.path.append("/home/stanley/work/python/ECCOv4-py/")
import ecco_v4_py as ecco

ecco.plot_tiles(
    z_omega - z_sigma,
    cmin=-200,
    cmax=200,
    fig_size=9,
    layout="latlon",
    rotate_to_latlon=True,
    Arctic_cap_tile_location=10,
    show_tile_labels=False,
    show_colorbar=True,
    show_cbar_label=True,
    cbar_label="Depth difference [m]",
)
plt.suptitle("z_omega - z_sigma")

plt.savefig("/home/stanley/Fig.png", bbox_inches="tight")

# In[Double check neutralocean's grid differencing]
from xmitgcm import open_mdsdataset
import xgcm

# To check that the grid works, we'll calculate the backwards differences of the
# sea-surface temperature in both horizontal directions. Begin by extracting SST.
k = 0  # index in vertical dimension
Tnp = T.values  # extract raw numpy data
SST = Tnp[..., k]  # slice the shallowest data from each water column

# Calculate difference between all pairs of adjacent casts
a, b = grid["edges"]
ΔT = SST.reshape(-1)[a] - SST.reshape(-1)[b]

# Decompose ΔT (a 1D array) into two 2D arrays, one for differences in each horizontal dimension
SSTx, SSTy = edgedata_to_maps(ΔT, n, face_connections, dims, xsh, ysh)


# Next, we'll repeat the above differencing using ECCO's methods.
# See https://ecco-v4-python-tutorial.readthedocs.io/VectorCalculus_ECCO_barotropicVorticity.html
print(
    "Download the grid's binary data files from "
    "< https://ndownloader.figshare.com/files/6494721 > "
    f" then exctract it so that {folder_ecco4} contains a folder 'global_oce_llc90'"
)
ds_llc = open_mdsdataset(
    folder_ecco4 + "global_oce_llc90/",
    iters=0,
    geometry="llc",
)

# rename 'face' to 'tile' to match T from the netCDF file
ds_llc = ds_llc.rename({"face": "tile"})

# Make the xgcm grid object to handle the differencing.
grid_llc = xgcm.Grid(
    ds_llc,
    periodic=False,
    face_connections=face_connections,
    coords={
        "X": {"center": "i", "left": "i_g"},
        "Y": {"center": "j", "left": "j_g"},
    },
)

# Backward differences in both horizontal directions
SSTx_ = grid_llc.diff(T[..., k], "X").values
SSTy_ = grid_llc.diff(T[..., k], "Y").values


# Begin checks.

# First, check the two differencing methods are equal, everywhere
assert np.array_equal(SSTx, SSTx_, equal_nan=True)
assert np.array_equal(SSTy, SSTy_, equal_nan=True)
print(
    "Test for equivalence between neutralocean grid differencing and xgcm grid differencing: passed."
)

# Now, check that differencing makes sense at an interior point
(t, j, i) = (9, 40, 40)
assert SSTx[t, j, i] == (Tnp[t, j, i, k] - Tnp[t, j, i - 1, k])
assert SSTy[t, j, i] == (Tnp[t, j, i, k] - Tnp[t, j - 1, i, k])
print(
    f"SSTx[{t}, {j}, {i}] == (T[{t}, {j}, {i}, {k}] - T[{t}, {j}, {i - 1}, {k}])"
)
print(
    f"SSTy[{t}, {j}, {i}] == (T[{t}, {j}, {i}, {k}] - T[{t}, {j - 1}, {i}, {k}])"
)

# Check that differencing makes sense across a boundary.  Find the second point
# that is involved in the difference across the boundary.  Verify by hand that
# this second point is where it ought to be.
# In this case, we're taking a meridional difference from the southern boundary
# of tile 4, and the second point is along the northern boundary of tile 3,
# which makes sense.
t, j, i = 4, 0, 40
idx = np.nonzero(SSTy[t, j, i] == (Tnp[t, j, i, k] - Tnp[..., k]))
t_, j_, i_ = idx[0][0], idx[1][0], idx[2][0]
assert SSTy[t, j, i] == Tnp[t, j, i, k] - Tnp[t_, j_, i_, k]
print(f"SSTy[{t},{j},{i}] == T[{t},{j},{i},{k}] - T[{t_},{j_},{i_},{k}]")
# # Output:
# SSTy[4,0,40] == T[4,0,40,0] - T[3,89,40,0]



"""Omega surfaces"""
import numpy as np
from time import time
from scipy.sparse import csc_matrix
from scipy.sparse.linalg import lsqr
from scipy.sparse.linalg import spsolve

from neutralocean.surface.isopycnal import _isopycnal
from neutralocean.surface._vertsolve import _make_vertsolve
from neutralocean.ppinterp import make_pp
from neutralocean.bfs import bfs_conncomp1, bfs_conncomp1_wet
from neutralocean.grid.graph import edges_to_graph
from neutralocean.ntp import ntp_epsilon_errors, ntp_epsilon_errors_norms
from neutralocean.lib import (
    xr_to_np,
    _xrs_in,
    _xr_out,
    _process_pin_cast,
    _process_casts,
    _process_eos,
    aggsum,
)
from neutralocean.mixed_layer import mixed_layer


def omega_surf(S, T, P, grid, **kwargs):
    """Calculate an omega surface from structured ocean data.

    Given 3D salinity, temperature, and pressure or depth data arranged on a
    rectilinear grid, calculate a 2D omega surface [1]_ [2]_, which is a
    highly accurate approximately neutral surface.

    Parameters
    ----------
    S, T, P : ndarray or xarray.DataArray

        See `potential_surf`

    p_init : ndarray, Default None

        Pressure or depth on the initial approximately neutral surface.

        See Examples section.

    ref : float, or tuple of float of length 2

        If `p_init` is None, the reference value(s) for the initial potential
        density surface or in-situ density (specific volume) anomaly surface
        that initializes the omega surface algorithm. If `ref` is a scalar, a
        potential density urface is used, and if `ref` is None, the reference
        `P` is `pin_p` (i.e. taken local to the pinning location). If `ref`
        is a tuple of length two, a in-situ density anomaly surface is used,
        and if `ref` is (None, None), then the reference `S` and `T` values
        are taken local to the pinning location (pressure or depth `pin_p` on
        the pinning cast `pin_cast`).

        See Examples section.

    isoval : float

        Isovalue for the initial potential density or in-situ density anomaly
        surface when `p_init` is not given.  Units are same as returned by
        the function specified by `eos`.

        See Examples section.

    pin_p, pin_cast :

        See `potential_surf`

    Returns
    -------
    s, t, p : ndarray or xarray.DataArray

        practical / Absolute salinity, potential / Conservative temperature,
        and pressure / depth on surface

    d : dict

        Diagnostics.
        The first four listed below give information going into the `i`'th
        iteration (e.g. the 0'th element is about the initial surface).
        The rest give information about what the `i`'th iteration did (and
        hence their 0'th elements are irrelevant).

        `"e_MAV"` : array of float

            Mean Absolute Value of the ϵ neutrality error on the surface,
            area-weighted.  Units are those of `eos` return values divided by
            those of `dist*` inputs.

        `"e_RMS"` : array of float

            As `"e_MAV"` but for the Root Mean Square.

        `"n_wet"`: array of float

            Number of wet casts (surface points).

        `"timer"` : array of float

            Time spent on each iteration, excluding set-up (approximately) and diagnostics.

        `"ϕ_MAV"` : array of float

            Mean Absolute Value of the Locally Referenced Potential Density
            perturbation, per iteration

        `"Δp_MAV"` : array of float

            Mean Absolute Value of the pressure or depth change from one
            iteration to the next

        `"Δp_RMS"` : array of float

            Root Mean Square of the pressure or depth change from one
            iteration to the next

        `"Δp_Linf"` : array of float

            Maximum absolute value (infinity norm) of the pressure or depth
            change from one iteration to the next

        `"n_newly_wet"` : array of int

            Number of casts that are newly wet, per iteration

        `"timer_bfs"` : array of float

            Time spent in Breadth-First Search including wetting, per iteration.

        `"timer_mat"` : array of float

            Time spent building and solving the matrix problem, per iteration.

        `"timer_update"` : array of float

            Time spent vertically updating the surface.

    Other Parameters
    ----------------
    grid, vert_dim, grav, rho_c, interp, diags, output, TOL_P_SOLVER :

        See `potential_surf`

    eos : str or tuple of functions, Default 'gsw'

        As in `potential_surf`, excluding the option to pass a single function.
        The omega surface algorithm relies on knowing the partial derivatives
        of the equation of state with respect to salinity and temperature, so
        the `eos_s_t` function is also required if given as a tuple of functions.

    ITER_MIN : int, Default 1

        Minimum number of iterations.

    ITER_MAX : int, Default 10

        Maximum number of iterations.

    ITER_START_WETTING : int, Default 1

        Iteration on which wetting begins.  Set to `np.inf` (`ITER_MAX` + 1
        would also do) to deactivate.

    ITER_STOP_WETTING : int, Default 5

        The last iteration on which to perform wetting.  This can be useful to
        avoid pesky water columns that repeatedly wet then dry.

    TOL_LRPD_MAV : float, Default 1e-7

        Exit iterations when the mean absolute value of the Locally Referenced
        Potential Density perturbation that updates the surface from one
        iteration to the next is less than this value.  Units are [kg m-3],
        even if `eos` returns a specific volume.  Set to 0 to deactivate.

    TOL_P_CHANGE_RMS : float, Default 0.0

        Exit iterations when the root mean square of the pressure or depth
        change on the surface from one iteration to the next is less than this
        value. Set to 0 to deactivate. Units are the same as `P` [dbar or m].

    p_ml : ndarray or dict, Default None

        If a dict, the pressure or depth at the base of the mixed layer is
        computed using `mixed_layer` with p_ml passed as keyword arguments,
        enabling control over the parameters in that function.
        See `mixed_layer` for details.

        If an ndarray (of the same shape as the lateral dimensions of `S`),
        the pressure or depth at the base of the mixed layer in each water
        column.

        When the surface's pressure is shallower than `p_ml` in any water
        column, it is set to NaN (a "dry" water column). This is not applied
        to the initial surface, but only to the surface after the first
        iteration, as the initial surface could be very far from neutral.

        If None, the mixed layer is not removed.

    OMEGA_FORMULATION : str, Default 'poisson'

        Specify how the matrix problem is set up and solved.  Options are
            - `'poisson'`, to solve the Poisson problem as in [1]_ with Cholesky, or
            - `'gradient'`, to solve the overdetermined gradient equations as in [2]_
              using LSQR.

    Examples
    --------
    omega surfaces require a pinning cast and initial surface.  The surface is
    iteratively updated while remaining fixed at the pinning cast.  The
    initial surface can be provided directly, as the surface with pressure or
    depth given by `p_init`, in the following method:

    >>> omega_surf(S, T, P, pin_cast, p_init, ...)

    Alternatively, a potential density surface or an in-situ density (specific
    volume) anomaly surface can be used as the initial surface.  To do this,
    use one of the following two methods:

    >>> omega_surf(S, T, P, ref, isoval, pin_cast, ...)

    >>> omega_surf(S, T, P, ref, pin_p, pin_cast, ...)

    For more info on these methods, see the Examples section of `potential_surf`.
    Note that `pin_cast` is always a required input.  Note that `ref` is
    needed to distinguish which of the two types of isopycnals will be used as
    the initial surface.

    Notes
    -----
    See `potential_surf` Notes.

    .. [1] Stanley, McDougall, Barker 2021, Algorithmic improvements to finding
       approximately neutral surfaces, Journal of Advances in Earth System
       Modelling, 13(5).

    .. [2] Klocker, McDougall, Jackett 2009, A new method of forming approximately
       neutral surfaces, Ocean Science, 5, 155-172.
    """

    ref = kwargs.get("ref")
    pin_p = kwargs.get("pin_p")
    pin_cast = kwargs.get("pin_cast")
    p_init = kwargs.get("p_init")
    vert_dim = kwargs.get("vert_dim", -1)
    p_ml = kwargs.get("p_ml")
    diags = kwargs.get("diags", True)
    output = kwargs.get("output", True)
    eos = kwargs.get("eos", "gsw")
    rho_c = kwargs.get("rho_c")
    grav = kwargs.get("grav")
    ITER_MIN = kwargs.get("ITER_MIN", 1)
    ITER_MAX = kwargs.get("ITER_MAX", 10)
    ITER_START_WETTING = kwargs.get("ITER_START_WETTING", 1)
    ITER_STOP_WETTING = kwargs.get("ITER_STOP_WETTING", 5)
    TOL_P_SOLVER = kwargs.get("TOL_P_SOLVER", 1e-4)
    TOL_LRPD_MAV = kwargs.get("TOL_LRPD_MAV", 1e-7)
    TOL_P_CHANGE_RMS = kwargs.get("TOL_P_CHANGE_RMS", 0.0)
    OMEGA_FORMULATION = kwargs.get("OMEGA_FORMULATION", "poisson")
    interp = kwargs.get("interp", "linear")

    # Build function that calculates coefficients of a piecewise polynomial
    # interpolant, doing 1 problem at a time, and knowing there will be no nans
    # in the input data.
    ppc_fn = make_pp(interp, kind="1", out="coeffs", nans=False)

    # Build function that evaluates two piecewise polynomial interpolants (with
    # the same set of independent data), doing n problems at a time.
    interp_two = make_pp(interp, kind="u", out="interp", num_dep_vars=2)

    sxr, txr, pxr = _xrs_in(S, T, P, vert_dim)  # before _process_casts
    pin_cast = _process_pin_cast(pin_cast, S)  # call before _process_casts
    S, T, P = _process_casts(S, T, P, vert_dim)
    eos, eos_s_t = _process_eos(eos, grav, rho_c, need_s_t=True)

    # Save shape of horizontal dimensions, then flatten horiz dims to 1D.
    surf_shape = S.shape[0:-1]
    N = np.prod(surf_shape)  # number of nodes (water columns)
    S, T, P = (np.reshape(X, (N, -1)) for X in (S, T, P))

    # Update pinning cast to a linear index, unless already so
    if isinstance(pin_cast, (tuple, list)):
        pin_cast = np.ravel_multi_index(pin_cast, surf_shape)

    # Prepare grid ratios for matrix problem.
    distratio = grid["distperp"] / grid["dist"]

    # Pre-calculate grid adjacency needed for Breadth First Search
    edges = grid["edges"]
    graph = edges_to_graph(edges, N)

    if OMEGA_FORMULATION.lower() == "poisson":
        global_solver = _omega_matsolve_poisson
    elif OMEGA_FORMULATION.lower() == "gradient":
        global_solver = _omega_matsolve_gradient
        distratio = np.sqrt(distratio)
    else:
        raise ValueError(
            f"Unknown OMEGA_FORMULATION. Given {OMEGA_FORMULATION}"
        )

    if eos(34.5, 3.0, 1000.0) < 1.0:
        # Convert from a density tolerance [kg m^-3] to a specific volume tolerance [m^3 kg^-1]
        TOL_LRPD_MAV = TOL_LRPD_MAV / 1000.0**2

    # Pre-allocate arrays for diagnostics
    if diags:
        d = {
            "e_MAV": np.zeros(ITER_MAX + 1, dtype=np.float64),
            "e_RMS": np.zeros(ITER_MAX + 1, dtype=np.float64),
            "timer": np.zeros(ITER_MAX + 1, dtype=np.float64),
            "ϕ_MAV": np.zeros(ITER_MAX + 1, dtype=np.float64),
            "Δp_MAV": np.zeros(ITER_MAX + 1, dtype=np.float64),
            "Δp_RMS": np.zeros(ITER_MAX + 1, dtype=np.float64),
            "Δp_Linf": np.zeros(ITER_MAX + 1, dtype=np.float64),
            "n_wet": np.zeros(ITER_MAX + 1, dtype=int),
            "n_newly_wet": np.zeros(ITER_MAX + 1, dtype=int),
            "timer_bfs": np.zeros(ITER_MAX + 1, dtype=np.float64),
            "timer_mat": np.zeros(ITER_MAX + 1, dtype=np.float64),
            "timer_update": np.zeros(ITER_MAX + 1, dtype=np.float64),
        }
    else:
        d = dict()

    timer = time()
    if p_init is None:
        # Calculate an initial "potential" or "anomaly" surface
        if isinstance(ref, (tuple, list)) and len(ref) == 2:
            ans_type = "anomaly"
        else:
            ans_type = "potential"

        # Update arguments with pre-processed values
        kwargs["vert_dim"] = -1  # Since S, T, P already reordered
        kwargs["diags"] = False  # Will make our own diags next
        kwargs["eos"] = eos
        kwargs["pin_cast"] = pin_cast  # update with the 1D value
        s, t, p, _ = _isopycnal(ans_type, S, T, P, **kwargs)

    else:
        # Handling and error checking on p_init
        p_init = xr_to_np(p_init)
        if not isinstance(p_init, np.ndarray):
            raise TypeError(
                'If provided, "p_init" or "p_init.values" must be an ndarray'
            )
        if p_init.shape != surf_shape:
            raise ValueError(
                f'"p_init" should contain a 2D array of size {surf_shape};'
                f" found size {p_init.shape}"
            )
        p_init = np.reshape(p_init, -1)  # now reshape to 1D

        if pin_p is not None and pin_p != p_init[pin_cast]:
            raise ValueError("pin_p does not match p_init at pin_cast")

        p = p_init.copy()

        # Interpolate S and T onto the surface
        # TODO: Update this to handle ice shelf cavity friendly interpolation
        s, t = interp_two(p, P, S, T, 0)

    pin_p = p[pin_cast]

    if np.isnan(p[pin_cast]):
        raise RuntimeError("The initial surface is NaN at the reference cast.")

    if ITER_MAX > 1 and isinstance(p_ml, dict):
        # Compute the mixed layer from parameter inputs
        p_ml = mixed_layer(S, T, P, eos, **p_ml)

    if p_ml is None:
        # Prepare array as needed for bfs_conncomp1_wet
        p_ml = np.full(p.shape, -np.inf)
        # p_ml = np.broadcast_to(-np.inf, (ni, nj))  # DEV: Doesn't work with @numba.njit
    else:
        # Ensure p_ml is 1D array, in case it was given as a 2D array.
        p_ml = np.reshape(p_ml, -1)

    # ensure same nan structure between s, t, and p. Just in case user gives
    # np.full((ni,nj), 1000) for a 1000dbar isobaric surface, for example
    p[np.isnan(s)] = np.nan

    if diags:
        d["timer"][0] = time() - timer

        e_RMS, e_MAV = ntp_epsilon_errors_norms(s, t, p, grid, eos_s_t)
        d["e_RMS"][0], d["e_MAV"][0] = e_RMS, e_MAV

        n_wet = np.sum(np.isfinite(p))
        d["n_wet"][0] = n_wet

        if output:
            print(
                "iter |"
                "    MAV(ϕ)     |"
                "    RMS(Δp)      |"
                " # wet casts (# new) |"
                "     RMS(ϵ)     |"
                " time (s)"
            )
            print(
                f"{0:4d} |"
                f"                                 |"
                f" {d['n_wet'][0]:11d}         |"
                f" {e_RMS:.8e} |"
                f" {d['timer'][0]:.3f}"
            )

    vertsolve = _make_vertsolve(eos, ppc_fn, "omega")

    # --- Begin iterations
    # Note: the surface exists wherever p is non-nan.  The nan structure of s
    # and t is made to match that of p when the vertical solve step is done.
    Δp_RMS = 0.0  # ensure this is defined; needed if TOL_P_CHANGE_RMS == 0
    for iter_ in range(1, ITER_MAX + 1):
        timer = time()

        # --- Remove the Mixed Layer
        if p_ml is not None and iter_ > 1:
            p[p < p_ml] = np.nan

        # --- Determine the connected component containing the reference cast, via Breadth First Search
        timer_loc = time()
        if iter_ >= ITER_START_WETTING and iter_ <= ITER_STOP_WETTING:
            bfsqu, n_newly_wet = bfs_conncomp1_wet(
                graph.indptr,
                graph.indices,
                pin_cast,
                s,
                t,
                p,
                S,
                T,
                P,
                TOL_P_SOLVER,
                eos,
                ppc_fn,
                p_ml,
            )
        else:
            bfsqu = bfs_conncomp1(
                graph.indptr, graph.indices, pin_cast, np.isfinite(p)
            )
            n_newly_wet = 0
        timer_bfs = time() - timer_loc

        # --- Solve global matrix problem for the exactly determined Poisson equation
        timer_loc = time()
        # Pre-sort the BFS queue: tests in both MATLAB and Python on OCCA data
        # show this gives an overall speedup for both Poisson and Gradient formulations.
        bfsqu = np.sort(bfsqu)
        ϕ = global_solver(s, t, p, edges, distratio, bfsqu, pin_cast, eos_s_t)
        timer_mat = time() - timer_loc

        # --- Update the surface (mutating s, t, p by vertsolve)
        timer_loc = time()
        p_old = p.copy()  # Record old surface for pinning and diagnostics
        vertsolve(s, t, p, S, T, P, ϕ, TOL_P_SOLVER)

        # DEV:  time seems indistinguishable from using factory function as above
        # _vertsolve_omega(s, t, p, S, T, P, Sppc, Tppc, ϕ, TOL_P_SOLVER, eos)

        # Force p to stay constant at the reference column, identically.
        # This avoids any intolerance from the vertical solver.
        p[pin_cast] = pin_p

        timer_update = time() - timer_loc

        # --- Closing Remarks
        ϕ_MAV = np.nanmean(abs(ϕ))
        if diags or TOL_P_CHANGE_RMS > 0:
            Δp = p - p_old
            Δp_RMS = np.sqrt(np.nanmean(Δp**2))

        if diags:

            d["timer"][iter_] = time() - timer

            Δp_MAV = np.nanmean(abs(Δp))
            Δp_Linf = np.nanmax(abs(Δp))

            # Diagnostics about what THIS iteration did
            d["ϕ_MAV"][iter_] = ϕ_MAV
            d["Δp_MAV"][iter_] = Δp_MAV
            d["Δp_RMS"][iter_] = Δp_RMS
            d["Δp_Linf"][iter_] = Δp_Linf
            d["n_newly_wet"][iter_] = n_newly_wet

            d["timer_mat"][iter_] = timer_mat
            d["timer_update"][iter_] = timer_update
            d["timer_bfs"][iter_] = timer_bfs

            # Diagnostics about the state AFTER this iteration
            e_RMS, e_MAV = ntp_epsilon_errors_norms(s, t, p, grid, eos_s_t)
            d["e_RMS"][iter_], d["e_MAV"][iter_] = e_RMS, e_MAV

            n_wet = np.sum(np.isfinite(p))
            d["n_wet"][iter_] = n_wet

            if output:
                print(
                    f"{iter_:4d} |"
                    f" {ϕ_MAV:.8e} |"
                    f" {Δp_RMS:.8e} |"
                    f" {n_wet:11d} ({n_newly_wet:5}) |"
                    f" {e_RMS:.8e} |"
                    f" {d['timer'][iter_]:.3f}"
                )

        # --- Check for convergence
        if (
            ϕ_MAV < TOL_LRPD_MAV or Δp_RMS < TOL_P_CHANGE_RMS
        ) and iter_ >= ITER_MIN:
            break

    if diags:
        # Trim diagnostics
        for k, v in d.items():
            d[k] = v[0 : iter_ + (k in ("e_MAV", "e_RMS"))]

    # Reshape (from 1D arrays) and put into DataArrays if appropriate
    s, t, p = (
        _xr_out(np.reshape(x, surf_shape), xxr)
        for (x, xxr) in ((s, sxr), (t, txr), (p, pxr))
    )

    return s, t, p, d


def _omega_matsolve_gradient(s, t, p, edges, sqrtdistratio, m, mref, eos_s_t):
    """Solve the Gradient formulation of the omega-surface global matrix problem

    Parameters
    ----------
    s, t, p, edges, m, mref, eos_s_t :
        See `_omega_matsolve_poisson`

    sqrtdistratio : array
        The square-root of the distance of the interface between adjacent
        water columns divided by the square-root of the distance between
        adjacent water columns.  That is, `np.sqrt(distperp / dist)`,
        where `distperp` and `dist` are as in the `geoemtry` input to
        `omega_surf`.

    Returns
    -------
    ϕ : ndarray
        See `_omega_matsolve_poisson`

    """
    N = len(m)  # Number of water columns
    ϕ = np.full(p.size, np.nan, dtype=p.dtype)

    # If there is only one water column, there are no equations to solve,
    # then the solution is simply phi = 0 at that water column, and nan elsewhere.
    # Note, N >= 1 should be guaranteed by omega_surf(), so N <= 1 should imply
    # N == 1.  If N >= 1 weren't guaranteed (m empty), this would throw an error.
    if N <= 1:  # There are definitely no equations to solve
        ϕ[m[0]] = 0.0  # Leave this isolated pixel at current pressure
        return ϕ.reshape(p.shape)

    a, b, e, fac, ref = _omega_matsolve_helper(
        s, t, p, edges, sqrtdistratio, m, mref, eos_s_t
    )

    rhs = np.concatenate((-e, [0.0]))  # add 0 for pinning equation

    # Build columns for matrix, including extra entry for pinning equation.
    # Note m[ref] is the reference cast, so the ref'th entry in the solution
    # vector of the matrix column corresponds to ϕ at the reference cast.
    c = np.concatenate((a, b, [ref]))

    # E = number rows in matrix. Round down ignores pinning equation
    E = len(c) // 2

    # r = [0, 1, ..., E-1, 0, 1, ..., E-1, E]
    r = np.concatenate((np.tile(np.arange(E), 2), [E]))

    # When distratio = 1, v is [1, 1, ..., 1, -1, -1, ..., -1, 1e-2]
    v = np.concatenate((-fac, fac, [1e-2]))

    mat = csc_matrix((v, (r, c)), shape=(E + 1, N))

    sol = lsqr(mat, rhs)[0]

    # Heave solution to be exactly 0 at pinning cast (to fix any intolerance
    # caused by lsqr converging before reaching the exact solution)
    sol -= sol[ref]

    ϕ[m] = sol

    ϕ = ϕ.reshape(p.shape)
    return ϕ


def _omega_matsolve_poisson(s, t, p, edges, distratio, m, mref, eos_s_t):
    """Solve the Poisson formulation of the omega-surface global matrix problem

    Parameters
    ----------
    s, t, p : ndarray

        Salinity, temperature, pressure on the surface

    edges : tuple of length 2 of 1D arrays

        See grid['edges'] from `omega_surf`

    distratio : array

        The distance of the interface between adjacent water columns divided by
        the distance between adjacent water columns, in the same order as `edges`.
        That is, `grid['distperp'] / grid['dist']` where `grid` is as input
        to `omega_surf`.

    m : array

        Linear indices to the nodes in order of the BFS, ie all casts in this
        connected component.
        That is, `m = qu[0:qt+1]` where `qu` and `qt` are outputs from
        `bfs_conncomp1` in bfs.py.

    mref : int

        Linear index to the reference cast, at which ϕ will be zero

    eos_s_t : function

        Function returning the partial derivatives of the equation of state
        with respect to S and T.

    Returns
    -------
    ϕ : ndarray

        Locally referenced potential density (LRPD) perturbation.
        Vertically heaving the surface so that its LRPD in water column m
        increases by the m'th element of ϕ will yield a more neutral surface.
    """
    N = len(m)  # number of water columns
    ϕ = np.full(p.size, np.nan, dtype=p.dtype)  # prealloc space

    # If there is only one water column, there are no equations to solve,
    # then the solution is simply phi = 0 at that water column, and nan elsewhere.
    # Note, N >= 1 should be guaranteed by omega_surf(), so N <= 1 should imply
    # N == 1.  If N >= 1 weren't guaranteed (m empty), this would throw an error.
    if N <= 1:  # There are definitely no equations to solve
        ϕ[m[0]] = 0.0  # Leave this isolated pixel at current pressure
        return ϕ.reshape(p.shape)

    # Get list of edges (a,b), ϵ neutrality errors, and geometric factors
    # for the current connected component containing the reference cast, and
    # map everything onto a set of N
    a, b, e, fac, ref = _omega_matsolve_helper(
        s, t, p, edges, distratio, m, mref, eos_s_t
    )

    # Divergence of ϵ,
    # D = ∑_{n ∈ N(m)} ϵₘₙ
    #   = ∑_{j=1}^E  δ_{m, aⱼ} ϵ_{m, bⱼ}  +  ∑_{j=1}^E  δ_{m, bⱼ} ϵ_{m, aⱼ}
    #   = ∑_{j=1}^E  δ_{m, aⱼ} ϵ_{m, bⱼ}  -  ∑_{j=1}^E  δ_{m, bⱼ} ϵ_{aⱼ, m}
    # achieved by
    D = aggsum(e, a, N) - aggsum(e, b, N)

    # Prepare diagonal entries of negative Laplacian.  For uniform geometry,
    # this simply counts the number of edges incident upon each node.  For
    # rectilinear grids, this value will be 4 for a typical node, but can be
    # less near boundaries of the connected component.
    diag = aggsum(fac, a, N) + aggsum(fac, b, N)

    # Build the rows, columns, and values of the sparse matrix
    r = np.concatenate((a, b, np.arange(N)))
    c = np.concatenate((b, a, np.arange(N)))
    v = np.concatenate((-fac, -fac, diag))  # negative Laplacian

    # Build the negative Laplacian sparse matrix with N rows and N columns
    L = csc_matrix((v, (r, c)), shape=(N, N))

    # Pinning surface at reference cast by ADDING the equation
    # 1 * ϕ[ref] = 0 to the ref'th equation.  Note, m[ref] is the reference cast.
    # If the BFS queue (m) were not sorted, then ref == 0 and m[0] would be the
    # reference cast, since the BFS is initialized from the reference cast.
    L[ref, ref] += 1

    # Alternative pinning strategy: change the mref'th equation to be 1 * ϕ[mref] = 0.
    # Then, since ϕ[mref] = 0, values of L in mref'th column are irrelevant, so set
    # these to zero to maintain symmetry.
    # Resulting output is same as above to many sig figs.
    # L[:, ref] = 0
    # L[ref, :] = 0
    # L[ref, ref] = 1
    # D[ref] = 0

    # Solve the matrix problem, L ϕ = D
    ϕ[m] = spsolve(L, D)

    # Fix machine precision errors by applying pinning condition exactly
    ϕ[mref] = 0.0

    ϕ = ϕ.reshape(p.shape)
    return ϕ


def _omega_matsolve_helper(s, t, p, edges, distratio, m, mref, eos_s_t):
    """
    Compute ϵ neutrality errors and geometry on the list of edges in the graph
    that are between pairs of "wet" casts.  Then, prune the list of edges to
    just these "wet" edges that are within the connected component containing
    the reference cast (whose linear index in the full space is `mref`), and
    remap all nodes (water columns) from being labelled (0, 1, ... ni*nj-1)
    for the entire 2D grid, to being labelled (0, 1, ..., N-1) where N is the
    number of wet casts in this connected component.  The connected component
    is determined by `m`, which is a list of nodes in the label set of the
    full space, (0, 1, ... ni*nj-1).
    """

    N = len(m)

    # `remap` changes from linear indices (0, 1, ..., ni*nj-1) for the entire
    # space (including land), into linear indices (0, 1, ..., N-1) for the
    # current connected component
    remap = np.full(p.size, -1, dtype=int)
    remap[m] = np.arange(N)

    # Select a subset of edges, namely those between two "wet" casts
    a, b = edges
    ge = (remap[a] >= 0) & (remap[b] >= 0)  # good edges
    a, b = a[ge], b[ge]

    e = ntp_epsilon_errors(s, t, p, (a, b), eos_s_t)

    if np.isscalar(distratio):
        fac = np.full(e.shape, distratio)
        if distratio != 1.0:
            e *= fac  # scale e
    else:
        fac = distratio[ge]
        e *= fac  # scale e

    # Henceforth we only refer to nodes in the connected component, so remap edges now
    a, b = remap[a], remap[b]

    # Also remap the index for the reference cast from 2D space to 1D list of wet casts
    ref = remap[mref]

    return a, b, e, fac, ref



import numpy as np


def synthocean(
    shape, pbot=4000.0, SSSSa=0.3, zonally_uniform=False, wrap=(False, False)
):
    """
    Synthetic idealization of the Pacific and Southern Ocean with tuneable parameters

    Parameters
    ----------
    shape : tuple of int
        A three element tuple giving the dimensions of the output.  The elements
        specify the number of points in longitude, latitude, and depth space,
        respectively.

    pbot : float, Default 4000.0
        Value of `P` at the bottommost data point, i.e. `P[-1]`

    SSSSa : float, Default 0.3
        Southern Sea Surface Salinity anomaly: the sea surface salinity is
        increased by `SSSS` in the southernmost grid cells, and by 0 in the
        northernmost grid cells, and linearly in between. The degree to
        which the southern casts are statically unstable can be controlled
        by this parameter. Increase to make more unstable.

    zonally_uniform : bool, Default False
        If `True`, the synthetic ocean is uniform in longitude, and the neutral
        helicity is zero.  If `False`, some zonal structure is created which
        results in non-zero neutral helicity.

    wrap : tuple of bool, Default (False, False)
        Specify periodicity of the lateral dimensions.
        When `wrap[0]` is `False`, `S[0, :, :] = T[0, :, :] = nan`.
        When `wrap[1]` is `False`, `S[:, 0, :] = T[:, 0, :] = nan`.


    Returns
    -------
    S, T : ndarray
        The practical / Absolute Salinity and potential / Conservative
        Temperature as a 3D array.

    P : ndarray
        The pressure / depth as a 1D array.

    g : dict
        The geometry of the grid, including latitude, longitude, cell widths
        and areas.  See code for details.

    """

    ni, nj, nk = shape

    X = np.linspace(-1, 1, ni).reshape((ni, 1, 1))  # longitude (temporarily scaled)
    Ynorth = 60
    Ysouth = -80
    Y = np.linspace(Ysouth, Ynorth, nj).reshape((1, nj, 1))  # latitude

    # Set surface T
    #   Linear model Poly5:
    #   f(x) = p1*x^5 + p2*x^4 + p3*x^3 + p4*x^2 + p5*x + p6
    #   where x is normalized by mean -7 and std 41.14
    #   Coefficients (with 95# confidence bounds):
    p1 = 0.6265  #  (0.3083, 0.9446)
    p2 = 3.269  #  (2.994, 3.545)
    p3 = -3.602  #  (-4.682, -2.522)
    p4 = -18.93  #  (-19.66, -18.19)
    p5 = 5.927  #  (5.108, 6.745)
    p6 = 27.41  #  (27.06, 27.76)
    x = (Y - -7) / 41.14
    Ts = p1 * x ** 5 + p2 * x ** 4 + p3 * x ** 3 + p4 * x ** 2 + p5 * x + p6

    if zonally_uniform:
        Ts = np.tile(Ts, (ni, 1, 1))  # no helicity
    else:
        # add some east-west structure
        Ts = Ts - 1.5 * X

    # contract extreme values into a given range
    Tmin = -1.8
    Tmax = 30.0
    Ts = (Ts - np.min(Ts)) / (np.max(Ts) - np.min(Ts)) * (Tmax - Tmin) + Tmin

    # Set surface S
    #   Linear model Poly4:
    #   f(x) = p1*x^4 + p2*x^3 + p3*x^2 + p4*x + p5
    #   where x is normalized by mean -7 and std 41.14
    #   Coefficients (with 95# confidence bounds):
    p1 = 0.2221  #  (0.1526, 0.2916)
    p2 = -0.09487  #  (-0.1553, -0.03444)
    p3 = -1.355  #  (-1.54, -1.17)
    p4 = -0.126  #  (-0.2438, -0.00815)
    p5 = 35.59  #  (35.51, 35.68)
    x = (Y - -7) / 41.14
    Ss = p1 * x ** 4 + p2 * x ** 3 + p3 * x ** 2 + p4 * x + p5

    if zonally_uniform:
        Ss = np.tile(Ss, (ni, 1, 1))  # no helicity
    else:
        # add some east-west structure (this adds helicity)
        Ss = Ss + (0.4 + 0.5 * X - 0.8 * X ** 2)

    # Tilt surface salinity function, so high SSS in South (if SSSSa > 0)
    Ss = Ss + np.linspace(SSSSa, 0, nj).reshape((1, nj, 1))

    # Set bottom T,S
    # linearly warmer and fresher moving northwards
    Tb = np.linspace(0, 1, nj).reshape((1, nj, 1)) + (Tmin - 0.1)
    Sb = np.linspace(34.7, 34.5, nj).reshape((1, nj, 1))

    # Add depth structure by linear interpolation of surface->bottom T,S data
    S = np.linspace(0, 1, nk).reshape((1, 1, nk)) * (Sb - Ss) + Ss
    T = np.linspace(0, 1, nk).reshape((1, 1, nk)) * (Tb - Ts) + Ts

    # Nonlinear spacing of pressure means dTdp and dSdp will be non-uniform
    P = np.linspace(0, 1, nk) ** 3 * pbot

    # Add walls
    if not wrap[0]:
        S[0, :, :] = np.nan
        T[0, :, :] = np.nan
    if not wrap[1]:
        S[:, 0, :] = np.nan
        T[:, 0, :] = np.nan

    X = (X + 1) * 180  # change longitude, to be [0, 360].

    g = dict()
    g["nx"] = ni  # number of zonal grid points
    g["ny"] = nj  # number of meridional grid points
    g["nz"] = nk  # number of vertical grid points
    g["rSphere"] = 6370000  # Earth radius [m]
    g["resx"] = ni / 360  # Number of tracer cells per zonal degree
    g["resy"] = nj / (Ynorth - Ysouth)  # Number of tracer cells per meridional degree

    # Override the loaded grid variables with ones calculated in accordance with
    # MITgcm's spherical polar grid: See INI_SPHERICAL_POLAR_GRID.F
    # You may wish to do this if your grid data is given in single precision.

    # Create our own spherical-polar grid. Short-cuts have been
    # taken in the formulas, appropriate for a uniform
    # latitude-longitude grid. See INI_SPHERICAL_POLAR_GRID.F
    # if adjustments must be made for non-uniform grid.
    deg2rad = np.pi / 180

    # fmt: off
    g["XGvec"] = 0 + np.arange(g["nx"]) / g["resx"]  # longitude on west edge of tracer cell
    g["YGvec"] = Ysouth + np.arange(g["ny"]) / g["resy"]  # latitude on south edge of tracer cell
    g["XCvec"] = g["XGvec"] + 1 / (2 * g["resx"]) # longitude at centre of tracer cell
    g["YCvec"] = g["YGvec"] + 1 / (2 * g["resy"]) # latitude at centre of tracer cell

    g["DXGvec"] = g["rSphere"] * np.cos(g["YGvec"] * deg2rad) / g["resx"] * deg2rad  # zonal distance of tracer cell at south edge [m]
    g["DYGsc"] = g["rSphere"] * deg2rad / g["resy"]  # meridional distance of tracer cell at west edge [m]
    g["DXCvec"] = g["rSphere"] * np.cos(g["YCvec"] * deg2rad) / g["resx"] * deg2rad  # zonal distance between cell centres of tracer cell and its western neighbour [m]
    g["DYCsc"] = g["DYGsc"]  # meridional distance between cell centres of tracer cell and its southern neighbour [m]

    g["RACvec"] = g["rSphere"] * g["rSphere"] / g["resx"] * deg2rad * abs( np.sin((g["YGvec"] + 1/g["resy"])*deg2rad) - np.sin(g["YGvec"]*deg2rad) )  # area of tracer cell [m2]
    g["RAWvec"] = g["RACvec"]  # area of cell centred on west edge of tracer cell [m2]
    g["RASvec"] = g["rSphere"] * g["rSphere"] / g["resx"] * deg2rad * abs( np.sin(g["YCvec"] * deg2rad) - np.sin((g["YCvec"] - 1/g["resy"])*deg2rad) )  # area of cell centred on south edge of tracer cell [m2]
    g["RAZvec"] = g["rSphere"] * g["rSphere"] / g["resx"] * deg2rad * abs( np.sin(g["YCvec"] * deg2rad) - np.sin((g["YCvec"] - 1/g["resy"])*deg2rad) )  # area of vorticity cell, centred on south-west corner of tracer cell [m2]
    # fmt: on

    return S, T, P, g



import numpy as np


def build_grid(dims, e1u=1.0, e2v=1.0, e2u=1.0, e1v=1.0):
    """
    Build pairs of adjacent grid points and associated geometry for a tripolar grid

    Parameters
    ----------
    dims : tuple of int
        Dimensions of the grid. Letting `dims == (nj, ni)`, then
        `nj` is the number of grid cells in the y (latitude-like) direction, &
        `ni` is the number of grid cells in the x (longitude-like) direction.

    e1u : float or 2D array, Default 1.0
        Distance in the x dimension that lives on the U cell.

    e2v : float or 2D array, Default 1.0
        Distance in the y dimension that lives on the V cell.

    e2u : float or 2D array, Default 1.0
        Distance in the y dimension that lives on the U cell.

    e1v : float or 2D array, Default 1.0
        Distance in the x dimension that lives on the V cell.

    Notes
    -----
    `e1u`, `e2v`, `e2u`, `e1v` can have shape `(nj,ni)` or `(nj+1, ni+2)`.
    The latter form has 1 cell of padding in the y dimension to handle the
    tripolar grid, and 1 cell of padding in the x dimension to handle the
    longitudinal periodicity. If given the latter form, this extra padding is
    removed internally. When of shape `(nj, ni)`, 

    - `e1u[j,i]` is the x distance on the U grid at [j,i+1/2],
    - `e2v[j,i]` is the y distance on the V grid at [j+1/2,i],
    - `e2u[j,i]` is the y distance on the U grid at [j,i+1/2],
    - `e1v[j,i]` is the x distance on the V grid at [j+1/2,i].

    Returns
    -------
    grid : dict
        Containing the following:

        edges : tuple of length 2
            Each element is an array of int of length `E`, where `E` is the number of
            edges in the grid's graph, i.e. the number of pairs of adjacent water
            columns (including land) in the grid.
            If `edges = (a, b)`, the nodes (water columns) whose linear indices are
            `a[i]` and `b[i]` are adjacent.

        dist : 1d array
            Horizontal distance between adjacent water columns (nodes).
            `dist[i]` is the distance between nodes whose linear indices are
            `edges[0][i]` and `edges[1][i]`.

        distperp : 1d array
            Horizontal distance of the face between adjacent water columns (nodes).
            `distperp[i]` is the distance of the interface between nodes whose
            linear indices are `edges[0][i]` and `edges[1][i]`.
    """
    grid = dict()
    grid["edges"] = _build_edges(dims)
    grid["dist"] = _build_edgedata(dims, (e2v, e1u))
    grid["distperp"] = _build_edgedata(dims, (e1v, e2u))
    return grid


def _build_edges(dims):
    """
    Build list of pairs of adjacent grid points, numbered 0, ..., N-1, on a
    tripolar grid (for ORCA)
    """

    assert len(dims) == 2, "Need `len(dims) == 2`, for a 2D horizontal grid."

    nj, ni = dims

    Ei = ni * nj  # number of edges in i dimension
    Ej = ni * nj  # number of edges in j dimension
    E = Ei + Ej  #  number of edges in total

    a = np.empty(E, dtype=int)  # prealloc space
    b = np.empty(E, dtype=int)

    # Build linear indices to the entire grid with one extra row of padding
    # for the northern latitudes. Do not add longitudinal padding.
    idx = np.arange((nj + 1) * ni).reshape((nj + 1, ni))

    # Handle tripolar grid's northern latitudes: using the extra row of padding,
    # the cell indexed (here using 0-indexing) by [nj - 1, i] is actually in the
    # domain (not the padded row) and is adjacent to [nj - 1, ni - i - 1]. So,
    # make the linear index at cell [nj, i] have the same value as the linear
    # index at cell [nj - 1, ni - i - 1].
    idx[-1, :] = np.flip(idx[-2, :])

    # Handle j dimension (non-periodic).
    # Cells [j, i] and [j+1, i] are adjacent, for 0 <= j <= nj-1. This is true
    # even for j == nj-1, thanks to the extra padded row and np.flip above.
    a[:Ej] = idx[:nj, :].reshape(-1)  # [j, i]
    b[:Ej] = idx[1:, :].reshape(-1)  # [j+1, i]

    # Handle i dimension (zonally periodic).
    # Cells [j, i] and [j, mod(i+1, ni)] are adjacent.
    a[Ej:] = idx[0:-1, :].reshape(-1)  # [j, i]
    b[Ej:] = np.roll(idx[0:-1, :], -1, axis=1).reshape(-1)  # [j, i+1]

    return a, b


def _build_edgedata(dims, data):
    """
    Build a 1D array of `data` in the same order as constructed by `_build_edges`

    Parameters
    ----------
    dims :
        See `build_grid`

    data : tuple of ndarray
        Data that lives on edges, i.e. between nodes (water columns).

        With `dims == (nj, ni)`, each element of `data` should be a 2D array of
        shape `(nj, ni)` --- but if they are of shape `(nj+1, ni+2)` they will
        be trimmed to remove the last row and the first and last columns.

        - `data[0][j, i]` lives between nodes [i, j] and [i, j+1], for `0 <= j <= nj-1`.

        - `data[1][j, i]` lives between nodes [i, j] and [i+1, j], for `0 <= i <= ni-1`.

        Example: `data = (e2v, e1u)` where `e1u` is the distance in the 1st (x)
        dimension on the U grid and `e2v` is the distance in the 2nd (y)
        dimension on the V grid. These can be of shape `(nj, ni)` or be
        of shape `(nj+1, ni+2)` as in the mesh_mask data file.

    Returns
    -------
    edgedata : array
        1D array of `data`.
        If `a, b = _build_edges(dims)`, then `edgedata[i]` is the
        value from `data` that corresponds to the interface between the grid
        cells indexed by `a[i]` and `b[i]`.
    """

    nj, ni = dims
    v, u = data
    if v.shape == (nj + 1, ni + 2):
        v = v[:-1, 1:-1]
    if u.shape == (nj + 1, ni + 2):
        u = u[:-1, 1:-1]

    return np.concatenate((v.reshape(-1), u.reshape(-1)))


def edgedata_to_maps(edgedata, dims):
    """
    Convert 1D array of data living on edges into two 2D arrays, one for each
    spatial dimension, for a tripolar grid

    Parameters
    ----------
    edgedata : array
        1D array of data that lives on edges of the grid's graph, given in the
        same order as the edges cosntructed from `_build_edges`.
        If `a, b = _build_edges(dims, periodic)`, then `edgedata[i]` is the
        data that lives at the interface between the grid cells indexed by
        `a[i]` and `b[i]`.

    dims :
        See `build_grid`

    Returns
    -------

    v, u : 2D array
        2D arrays of shape `dims`. `v` and `u` contains the values from `edgedata`
        that correspond to data living on the V and U grids, respectively.
    """

    nj, ni = dims
    Ej = ni * nj  # number of edges in j dimension
    v = np.reshape(edgedata[:Ej], dims)
    u = np.reshape(edgedata[Ej:], dims)

    return v, u



"""
Density of Sea Water using the TEOS-10 Gibbs Seawater [1]_ function

Functions:

rho :: computes in-situ density from salinity, potential temperature and
    pressure

rho_s_t :: compute the partial derivatives of in-situ density with
    respect to salinity and potential temperature

rho_p :: compute the partial derivative of in-situ density with
    respect to pressure

Notes:
To make Boussinesq versions of these functions, see 
`neutralocean.eos.tools.make_eos_bsq`.

To make vectorized versions of these functions, see
`neutralocean.eos.tools.vectorize_eos`.

.. [1] McDougall, T.J. and P.M. Barker, 2011: Getting started with TEOS-10 and 
the Gibbs Seawater (GSW) Oceanographic Toolbox, 28pp., SCOR/IAPSO WG127, 
ISBN 978-0-646-55621-5. 
"""

import ctypes
import numba as nb
import gsw


# The following shows how we can access the C library scalar functions that
# are used by GSW-Python, since it is much faster for our jit functions to
# go straight to them rather than going via the Python ufunc wrappers.
dllname = gsw._gsw_ufuncs.__file__
gswlib = ctypes.cdll.LoadLibrary(dllname)
rho_gsw_ctypes = gswlib.gsw_rho  # In-situ density.
rho_gsw_ctypes.restype = ctypes.c_double
rho_gsw_ctypes.argtypes = (ctypes.c_double, ctypes.c_double, ctypes.c_double)


# Wrapping the ctypes function with njit reduces call overhead and makes it
# hashable, which is required for the caching we do via the lru_cache
# decorator.
@nb.njit(nb.f8(nb.f8, nb.f8, nb.f8))
def rho(s, t, p):
    """
    The TEOS-10 GSW equation of state for in-situ density.

    Parameters
    ----------
    s : float
        Absolute Salinity [g kg-1]
    t : float
        Conservative Temperature [degC]
    p : float
        Pressure [dbar]

    Returns
    -------
    rho : float
        In-situ density [kg m-3]

    """
    return rho_gsw_ctypes(s, t, p)


# Can't use nb.njit on functions wrapping gsw.rho_first_derivatives:
# "Unknown attribute 'rho_first_derivatives' of type Module(<module 'gsw' from ..."
# For the partial derivatives, we only call these on the entire surface,
# so the Python ufunc wrappers are okay.
def rho_s_t(s, t, p):
    """
    The partial derivatives of the TEOS-10 GSW equation of state for in-situ
    density with respect to Absolute Salinity and Conservative Temperature

    Parameters
    ----------
    s, t, p : float or ndarray
        See `rho`

    Returns
    -------
    rho_s : float or ndarray
        Partial derivative of in-situ density with respect to Absolute
        Salinity [kg m-3 (g/kg)-1]

    rho_t : float or ndarray
        Partial derivative of in-situ density with respect to Conservative
        Temperature [kg m-3 degC-1]

    """
    # Would be better if we didn't have to compute the pressure derivative.
    rs, rt, _ = gsw.rho_first_derivatives(s, t, p)
    return rs, rt


def rho_p(s, t, p):
    """
    The partial derivatives of the TEOS-10 GSW equation of state for in-situ
    density with respect to pressure

    Parameters
    ----------
    s, t, p : float or ndarray
        See `rho`

    Returns
    -------
    rho_p : float or ndarray
        Partial derivative of in-situ density with respect to pressure [kg m-3 dbar-1]

    """
    # Would be better if we didn't have to compute the S and T derivative.
    _, _, rp = gsw.rho_first_derivatives(s, t, p)
    return rp * 1e4  # convert from [kg m-3 Pa-1] to [kg m-3 dbar-1]



"""
Calculate approximately neutral surfaces that are isosurfaces of an function
with a closed form mathematical expression, namely 
potential density (or specific volume) surfaces and 
in-situ density (or specific volume) anomaly surfaces.
"""

import numpy as np
from time import time

from neutralocean.surface._vertsolve import _make_vertsolve
from neutralocean.ppinterp import ppval_1_two, make_pp, valid_range_1
from neutralocean.ntp import ntp_epsilon_errors_norms
from neutralocean.lib import (
    _xrs_in,
    _xr_out,
    _process_pin_cast,
    _process_casts,
    _process_eos,
)


def potential_surf(S, T, P, **kwargs):
    """Calculate a potential density (or specific volume) surface.

    Given practical / Absolute salinity `S`, potential / Conservative
    temperature `T`, and pressure (when non-Boussinesq) or depth
    (when Boussinesq) `P`, and given a reference pressure / depth `ref`,
    calculate an isosurface of `eos(S, T, ref)` where `eos` is the equation
    of state.

    Parameters
    ----------
    S, T : ndarray or xarray.DataArray

        3D practical / Absolute salinity and potential / Conservative
        temperature --- that is, a 2D array of 1D water columns or "casts"

    P : ndarray or xarray.DataArray

        In the non-Boussinesq case, `P` is the 3D pressure, sharing the same
        dimensions as `S` and `T`.

        In the Boussinesq case, `P` is the depth and can be 3D with the same
        structure as `S` and `T`, or can be 1D with as many elements as there
        are in the vertical dimension of `S` and `T`.

    ref : float

        The reference pressure or depth, in the same units as P.

        If `ref` is None, the reference value is taken as `pin_p`.

        See Examples section.

    isoval : float

        Isovalue of the potential density surface.
        Units are same as returned by `eos`.

        See Examples section.

    pin_p : float

        Pressure or depth at which the surface intersects the cast `pin_cast`.

        See Examples section.

    pin_cast : tuple or list of int of length 2

        Index for cast where surface is at pressure or depth `pin_p`.

        See Examples section.

    Returns
    -------
    s, t, p : ndarray or xarray.DataArray

        practical / Absolute salinity, potential / Conservative temperature,
        and pressure / depth on the surface

    d : dict

        Diagnostics.

        `"e_MAV"` : float

            Mean Absolute Value of the ϵ neutrality error on the surface,
            area-weighted.  Units are those of `eos` return values divided by
            those of `dist*` inputs.

        `"e_RMS"` : float

            As `"e_MAV"` but for the Root Mean Square.

        `"n_wet"`: float

            Number of wet casts (surface points).

        `"timer"` : float

            Time spent on the whole algorithm, excluding set-up and diagnostics.

        `"ref"` : float

            Reference pressure / depth for surface (matching input `ref` if given,
            otherwise this is calculated internally).

        `"isoval"` : float

            Isovalue of potential density for surface (matching input `isoval`
            if given, otherwise this is calculated internally).

    Other Parameters
    ----------------
    grid : dict
        Containing the following:

        edges : tuple of length 2
            Each element is an array of int of length E, where E is the number of
            edges in the grid's graph, i.e. the number of pairs of adjacent water
            columns (including land) in the grid.
            If `edges = (a, b)`, the nodes (water columns) whose linear indices are
            `a[i]` and `b[i]` are adjacent.
            Required if `diags` is True

        dist : 1d array
            Horizontal distance between adjacent water columns (nodes).
            `dist[i]` is the distance between nodes whose linear indices are
            `edges[0][i]` and `edges[1][i]`.
            If absent, a value of 1.0 is assumed for all edges.

        distperp : 1d array
            Horizontal distance of the face between adjacent water columns (nodes).
            `distperp[i]` is the distance of the interface between nodes whose
            linear indices are `edges[0][i]` and `edges[1][i]`.
            If absent, a value of 1.0 is assumed for all edges.

        For a rectilinear grid (e.g. latitude-longitude), use
            `neutralocean.grid.rectilinear.build_grid`

        For a tiled rectilinear grid, such as works with XGCM, use
            `neutralocean.grid.xgcm.build_grid`

        For a general grid given as a graph, use
            `neutralocean.grid.graph.build_grid`

        Also see the examples in `neutralocean.examples`.

    vert_dim : int or str, Default -1

        Specifies which dimension of `S`, `T` (and `P` if 3D) is vertical.

        If `S` and `T` are `ndarray`, then `vert_dim` is the `int` indexing
        the vertical dimension of `S` and `T` (e.g. -1 indexes the last
        dimension).

        If `S` and `T` are `xarray.DataArray`, then `vert_dim` is a `str`
        naming the vertical dimension of `S` and `T`.

        Ideally, `vert_dim` is -1.  See `Notes`.

    eos : str or function or tuple of functions, Default 'gsw'

        Specification for the equation of state.

        If a str, can be any of the strings accepted by
        `neutralocean.eos.tools.make_eos`,
        e.g. `'jmd95'`, `'jmdfwg06'`, `'gsw'`.

        If a function, must take three inputs corresponding to `S`, `T`, and
        `P`, and output the density (or specific volume).  This form is not
        allowed when `diags` is `True`.  This can be made as, e.g.,
        `eos = neutralocean.eos.make_eos('gsw')`
        for a non-Boussinesq ocean, or as
        `eos = neutralocean.eos.make_eos('gsw', grav, rho_c)`
        for a Boussinesq ocean with `grav` and `rho_c` (see inputs below).

        If a tuple of functions, the first element must be a function for the
        equation of state as above, and the second element must be a function
        taking the same three inputs as above and returning two outputs, namely
        the partial derivatives of the equation of state with respect to `S`
        and `T`.  The second element can be made as, e.g.,
        `eos_s_t = neutralocean.eos.make_eos_s_t('gsw', grav, rho_c)`

        The function (or the first element of the tuple of functions) should be
        `@numba.njit` decorated and need not be vectorized -- it will be called
        many times with scalar inputs.

    grav : float, Default None
        Gravitational acceleration [m s-2].  When non-Boussinesq, pass `None`.

    rho_c : float, Default None
        Boussinesq reference density [kg m-3].  When non-Boussinesq, pass `None`.

    interp : str, Default 'linear'

        Method for vertical interpolation.  Use `'linear'` for linear
        interpolation, and `'pchip'` for Piecewise Cubic Hermite Interpolating
        Polynomials.  Other interpolants can be added through the subpackage,
        `ppinterp`.

    diags : bool, Default True

        If True, calculate diagnostics (4th output).  If False, 4th output is
        an empty dict.

    output : bool, Default True

        If `True`, prints diagnostic output during computation.
        `diags` must be `True` for this to have any effect.
        To redirect this output to a file, do the following
        >>> import sys
        >>> tmp = sys.stdout
        >>> sys.stdout = file_id = open('myfile.txt', 'w')
        >>> # Now call this function ...
        >>> sys.stdout = tmp
        >>> file_id.close()

    TOL_P_SOLVER : float, Default 1e-4

        Error tolerance when root-finding to update the pressure or depth of
        the surface in each water column. Units are the same as `P`.

    Examples
    --------
    The output surface must be specified by some combination of reference
    pressure / depth `ref`, isovalue `isoval`, pinning cast `pin_cast` and
    pinning pressure `pin_p`.  The following methods are valid, and listed in
    order of precedence (e.g. `pin_cast` and `pin_p` are not used if both
    `ref` and `isoval` are given).

    >>> potential_surf(S, T, P, ref=___, isoval=___, ...)

    This finds the surface with given reference pressure / depth and the given
    isovalue.

    >>> potential_surf(S, T, P, ref=___, pin_p=___, pin_cast=___, ...)

    This finds the surface with given reference pressure / depth that intersects
    the given cast at the given pressure or depth.

    >>> potential_surf(S, T, P, pin_p=___, pin_cast=___, ...)

    This is as for the previous method, but selects the reference pressure /
    depth as `pin_p` (i.e. the local `P` value at the given cast's given
    pressure or depth).

    Notes
    -----
    This code will internally re-arrange `S`, `T`, `P` to have the vertical
    dimension last, so that the data for an individual water column is
    contiguous in memory.  If you call this function many times, consider
    using `lib._process_casts` to pre-process your `S`, `T`, `P` inputs to
    have the vertical dimension last.
    """

    return _isopycnal("potential", S, T, P, **kwargs)


def anomaly_surf(S, T, P, **kwargs):
    """Calculate a specific volume (or in-situ density) anomaly surface.

    Given practical / Absolute salinity `S`, potential / Conservative
    temperature `T`, and pressure / depth `P`, and given reference values
    `S0` and `T0`, calculate  an isosurface of `eos(S, T, P) - eos(S0, T0, P)`
    where `eos` is the equation of state.

    In a non-Boussinesq ocean, `P` is pressure.  Also, if one is computing
    geostrophic streamfunctions, it is most convenient if `eos` provides the
    specific volume.

    In a Boussinesq ocean, `P` is depth.  Also, if one is computing
    geostrophic streamfunctions, it is most convenient if `eos` provides the
    in-situ density.

    Parameters
    ----------
    S, T, P : ndarray or xarray.DataArray
        See `potential_surf`

    ref : tuple of float of length 2

        The reference S and T values.

        If `ref` is None or has a None element, the reference values are taken
        from the local `S` and `T` at the pressure or depth `pin_p` on the
        pinning cast `pin_cast`.

    isoval, pin_p, pin_cast :
        See `potential_surf`

    Returns
    -------
    s, t, p, d :
        See `potential_surf`.
        Note `d["ref"]` returns a 2 element tuple, namely `ref` as here.

    Other Parameters
    ----------------
    grid, vert_dim, eos, grav, rho_c, interp, diags, output, TOL_P_SOLVER :
        See `potential_surf`

    Examples
    --------
    The output surface must be specified by some combination of reference
    salinity and temperature `ref`, isovalue `isoval`, pinning cast `pin_cast` and
    pinning pressure `pin_p`.  The following methods are valid, and listed in
    order of precedence (e.g. `pin_cast` and `pin_p` are not used if both
    `ref` and `isoval` are given).

    >>> anomaly_surf(S, T, P, ref=___, isoval=___, ...)

    This finds the surface with given reference salinity and temperature and
    the given isovalue.

    >>> anomaly_surf(S, T, P, ref=___, pin_p=___, pin_cast=___, ...)

    This finds the surface with given reference salinity and temperature that
    intersects the given cast at the given pressure or depth.

    >>> anomaly_surf(S, T, P, pin_p=___, pin_cast=___, ...)

    This is as for the previous method, but selects the reference salinity and
    temperature from the local `S` and `T` values at the given cast's given
    pressure or depth.

    Notes
    -----
    See `potential_surf`.
    """

    return _isopycnal("anomaly", S, T, P, **kwargs)


def _isopycnal(ans_type, S, T, P, **kwargs):
    """Calculate an isosurface of potential density or specific volume anomaly.

    Inputs are as in `potential_surf` and `anomaly_surf`, but first input is a
    string specifying "potential" or "anomaly" """

    ref = kwargs.get("ref")
    isoval = kwargs.get("isoval")
    pin_cast = kwargs.get("pin_cast")
    pin_p = kwargs.get("pin_p")
    vert_dim = kwargs.get("vert_dim", -1)
    TOL_P_SOLVER = kwargs.get("TOL_P_SOLVER", 1e-4)
    eos = kwargs.get("eos", "gsw")
    rho_c = kwargs.get("rho_c")
    grav = kwargs.get("grav")
    diags = kwargs.get("diags", True)
    output = kwargs.get("output", True)
    grid = kwargs.get("grid")
    interp = kwargs.get("interp", "linear")

    # Build function that calculates coefficients of a piecewise polynomial
    # interpolant, doing 1 problem at a time, and knowing there will be no nans
    # in the input data.
    ppc_fn = make_pp(interp, kind="1", out="coeffs", nans=False)

    # Process arguments
    sxr, txr, pxr = _xrs_in(S, T, P, vert_dim)  # before _process_casts
    pin_cast = _process_pin_cast(pin_cast, S)  # call before _process_casts
    S, T, P = _process_casts(S, T, P, vert_dim)
    eos, eos_s_t = _process_eos(eos, grav, rho_c, need_s_t=diags)
    if diags and not (isinstance(grid, dict) and "edges" in grid):
        raise ValueError("grid['edges'] must be provided when diags is True")

    # Error checking on (ref, isoval, pin_cast, pin_p), then convert this
    # selection to (ref, isoval) pair
    _check_ref(ans_type, ref, isoval, pin_cast, pin_p, S)
    ref, isoval = _choose_ref_isoval(
        ans_type, ref, isoval, pin_cast, pin_p, eos, S, T, P, ppc_fn
    )

    # Solve non-linear root finding problem in each cast
    vertsolve = _make_vertsolve(eos, ppc_fn, ans_type)
    timer = time()
    s, t, p = vertsolve(S, T, P, ref, isoval, TOL_P_SOLVER)

    if pin_p is not None:  # pin_cast must also be valid
        # Adjust the surface at the pinning cast slightly, to match the pinning
        # pressure / depth.  This fixes small deviations of order `TOL_P_SOLVER`
        n = pin_cast
        p[n] = pin_p
        Sn, Tn, Pn = S[n], T[n], P[n]
        k, K = valid_range_1(Sn + Pn)  # Sn and Tn have same nan-structure
        Sppcn = ppc_fn(Pn[k:K], Sn[k:K])
        Tppcn = ppc_fn(Pn[k:K], Tn[k:K])
        s[n], t[n] = ppval_1_two(pin_p, Pn[k:K], Sppcn, Tppcn)

    d = dict()
    d["ref"] = ref
    d["isoval"] = isoval
    if diags:
        d["timer"] = time() - timer
        e_RMS, e_MAV = ntp_epsilon_errors_norms(s, t, p, grid, eos_s_t)
        d["e_RMS"], d["e_MAV"] = e_RMS, e_MAV

        n_wet = np.sum(np.isfinite(p))
        d["n_wet"] = n_wet
        if output:
            print(
                f"{ans_type} done"
                f" | {n_wet:11d} wet casts"
                f" | RMS(ϵ) = {e_RMS:.8e}",
                f" | {d['timer']:.3f} sec",
            )

    s, t, p = (_xr_out(x, xxr) for (x, xxr) in ((s, sxr), (t, txr), (p, pxr)))
    return s, t, p, d


def _check_ref(ans_type, ref, isoval, pin_cast, pin_p, S):
    """Error checking on ref / isoval / pin_cast / pin_p combinations for "potential"
    and "anomaly" surfaces
    """
    # First check None values to validate one of the following options:
    # >>> _isopycnal(ans_type, S, T, P, ref, isoval)
    # >>> _isopycnal(ans_type, S, T, P, ref, pin_cast, pin_p)
    # >>> _isopycnal(ans_type, S, T, P, pin_cast, pin_p)
    if ref is None:
        if pin_cast is None or pin_p is None:
            raise TypeError(
                'If "ref" is not provided, "pin_cast" and "pin_p" must be provided'
            )
    else:  # ref is not None
        if isoval is None and (pin_cast is None or pin_p is None):
            raise TypeError(
                'If "ref" is provided, either "isoval" must be provided or'
                ' "pin_cast" and "pin_p" must be provided'
            )

    # Error checking on ref
    if ref is not None:
        if ans_type == "potential":
            if not isinstance(ref, float):
                raise TypeError(
                    'For "potential" surfaces, if provided "ref" must be a float'
                )
        else:  # ans_type == "anomaly"
            if not (isinstance(ref, (tuple, list)) and len(ref) == 2):
                raise TypeError(
                    'For "anomaly" surfaces, if provided "ref" must be 2 element tuple/list of float'
                )

    # Error checking on pin_cast.  Let dict inputs (for xarray) pass through fine...
    if not isinstance(pin_cast, (type(None), dict)):
        try:
            S[pin_cast]
        except:
            raise ValueError(
                'If provided, "pin_cast" must be able to index all but the'
                " vertical dimension of S"
            )

    # Error checking on pin_p
    if not isinstance(pin_p, (type(None), float)):
        raise TypeError('If provided, "pin_p" must be a float')


def _choose_ref_isoval(
    ans_type, ref, isoval, pin_cast, pin_p, eos, S, T, P, ppc_fn
):
    # Handle the three valid calls in the following order of precedence:
    # >>> _isopycnal(ans_type, S, T, P, ref, isoval)
    # >>> _isopycnal(ans_type, S, T, P, ref, pin_cast, pin_p)
    # >>> _isopycnal(ans_type, S, T, P, pin_cast, pin_p)
    if isoval is None:  # => pin_cast and pin_p are both not None
        n = pin_cast  # evaluate S and T on the surface at the chosen location
        # s0, t0 = interp_fn(pin_p, P[n], S[n], T[n])

        Sn, Tn, Pn = S[n], T[n], P[n]
        k, K = valid_range_1(Sn + Pn)  # Sn and Tn have same nan-structure
        Sppcn = ppc_fn(Pn[k:K], Sn[k:K])
        Tppcn = ppc_fn(Pn[k:K], Tn[k:K])
        s0, t0 = ppval_1_two(pin_p, Pn[k:K], Sppcn, Tppcn)

        if ans_type == "potential":
            if ref is None:
                ref = pin_p
            isoval = eos(s0, t0, ref)
        else:  # ans_type == "anomaly"
            if ref is None or any(x is None for x in ref):
                ref = (s0, t0)
            isoval = eos(s0, t0, pin_p) - eos(
                ref[0], ref[1], pin_p
            )  #  == 0 when ref = (s0, t0)

    return ref, isoval



import numpy as np
import mpmath as mp
from scipy.sparse import coo_matrix, csr_matrix

from neutralocean.lib import aggsum
from neutralocean.bfs import bfs_conncomp1
from neutralocean.grid.graph import edges_to_graph
from neutralocean.ppinterp import make_pp

# def _linprob(s, t, p, ds, dt, dds, ddt, M, grid, eos, vc, DERIV):
def omegapp(s, t, p, M, grid, eos_s_t):
    """Calculate the Jacobian and Hessian of E"""
    N = len(M)

    # `remap` changes from linear indices (0, 1, ..., ni*nj-1) for the entire
    # space (including land), into linear indices (0, 1, ..., N-1) for the
    # current connected component
    remap = np.full(p.size, -1, dtype=int)
    remap[M] = np.arange(N)

    # Select a subset of edges, namely those between two "wet" casts that are
    # in this connected component.
    a, b = grid["edges"]
    ge = (remap[a] >= 0) & (remap[b] >= 0)  # good edges
    a, b = a[ge], b[ge]

    fac = grid["distratio"][ge]
    dperp = grid["distperp"][ge]
    areaG = grid["dist"][ge] * dperp
    
    sa, ta, pa = (x[a] for x in (s, t, p))
    sb, tb, pb = (x[b] for x in (s, t, p))

    rs, rt = eos_s_t(0.5 * (sa + sb), 0.5 * (ta + tb), 0.5 * (pa + pb))
    
    Δs = sb - sa
    Δt = tb - ta
    e = (rs * Δs + rt * Δt)
    
    B = e * 100

    # e = ntp_epsilon_errors(s, t, p, (a, b), eos_d3)
    # sa, ta, pa = (graph_binary_fcn(edges, x, avg1) for x in (s, t, p))
    # Δs, Δt = (graph_binary_fcn(edges, x, dif1) for x in (s, t))

    Hab = fac
    Hba = fac

    # Henceforth we only refer to nodes in the connected component, so remap edges now
    a, b = remap[a], remap[b]

    # For omega, the matrix is the negative Laplacian.  For uniform geometry,
    # this simply counts the number of edges incident upon each node.  For
    # rectilinear grids, this value will be 4 for a typical node, but can be
    # less near boundaries of the connected component.  This is achieved by
    #   diag = aggsum(fac, a, N) + aggsum(fac, b, N)
    # and then using (-fac) on the off-diagonal.
    # For ONS, the matrix is the Hessian of E.
    # Diagonal of Hessian is
    # diag[m] = 
    #   = ∑_{n ∈ N(m)} Bₘₙ
    #   =   ∑_{j=1}^E  δ_{m, aⱼ} B_{m, bⱼ} + ∑_{j=1}^E  δ_{m, bⱼ} B_{m, aⱼ}
    #   =   ∑_{j=1}^E  δ_{m, aⱼ} B_{m, bⱼ} - ∑_{j=1}^E  δ_{m, bⱼ} B_{aⱼ, m}
    #   =   np.sum(Bab[a == m]) - np.sum(Bab[b == m])
    #   =   aggsum(Bab, a, N)[m] - aggsum(Bab, b, N)[m]
    diag = (
        aggsum(-fac + B * dperp - 0.5 * B**2 * areaG, a, N) +
        aggsum(-fac - B * dperp - 0.5 * B**2 * areaG, b, N)
    )

    # Build the rows, columns, and values of the sparse matrix
    r = np.concatenate((a, b, np.arange(N)))
    c = np.concatenate((b, a, np.arange(N)))
    v = np.concatenate((Hab, Hba, diag))

    return v, r, c


# Pre-calculate grid adjacency needed for Breadth First Search
p = z_omega
P = Z
M0 = I0

grid1 = grid.copy()
grid1["dist"][:] = 1.0
grid1["distperp"][:] = 1.0
grid1["distratio"][:] = 1.0

interp_name = "pchip"
I2 = make_pp(interp_name, "u", out="interp", num_dep_vars=2)
s, t = I2(p, P, S, T, 0)

N = np.prod(p.shape)  # number of nodes (water columns)
edges = grid1["edges"]
graph = edges_to_graph(edges, N)
good = np.isfinite(mp.fp.matrix(p))
M = bfs_conncomp1(graph.indptr, graph.indices, M0, good)
v, r, c = omegapp(s, t, p, M, grid1, eos_s_t)

Hmp = mp.matrix(len(M), len(M))
for i in range(len(r)):
    Hmp[r[i], c[i]] += v[i]

Hnp = csr_matrix((np.float64(v), (r, c)), shape=(len(M), len(M)))
    
Es, Qs = mp.eigsy(Hmp)
Esnp = np.array(mp.fp.matrix(Es))  # convert to np.array of floats
i = int(np.argmin(abs(Esnp)))  # index to smallest abs eigenvalue
eigval = Es[i]
eigvec = Qs[:, i]

# print(Hnp[:,5].toarray())
print(eigval)


import numpy as np
import xarray as xr

import pooch


def load_OCCA():
    """
    Load grid, salinity, and potential temperature for OCCA 2004-2006 average

    Returns
    -------
    g : dict
        Grid information

    S : xarray.DataArray
        Salinity

    T : xarray.DataArray
        Potential temperature
    """

    # Use friendly pooch to download dataset, if it hasn't already done so
    # and cached it locally.
    url_salt = "https://dataverse.harvard.edu/api/access/datafile/:persistentId?persistentId=doi:10.7910/DVN/RNXA2A/VREV1Z"
    url_theta = "https://dataverse.harvard.edu/api/access/datafile/:persistentId?persistentId=doi:10.7910/DVN/RNXA2A/APHURG"

    hash_salt = (
        "184ca879e0ed18a078f31160c16ec078701e0f1af4d6de44bd1e70cc200d111c"
    )
    hash_theta = (
        "7fcef4663439e34b13de4b08a373a0e17b83bee2c621ea8ac07a8175fe15368e"
    )

    file_salt = pooch.retrieve(url=url_salt, known_hash=hash_salt)
    file_theta = pooch.retrieve(url=url_theta, known_hash=hash_theta)

    ts = 0  # Select the first time step (January)

    # Read grid info from the theta nc file
    ds = xr.open_dataset(file_theta).load()

    # Build our own grid, as the MITgcm does it
    # Pa2db = 1e-4
    d2r = np.pi / 180

    g = dict()  # model grid and parameters
    g["rho_c"] = 1027.5  # A guess. Same as ECCO2
    g["grav"] = 9.81  # A guess. Same as ECCO2
    g["rSphere"] = 6.37e6  # A guess. Same as ECCO2
    g["resx"] = 1  # 1 grid cell per zonal degree
    g["resy"] = 1  # 1 grid cell per meridional degree
    g["wrap"] = (True, False)  # periodic in longitude, not in latitude

    # Lateral coordinates
    g["XCvec"], g["XGvec"], g["YCvec"], g["YGvec"] = (
        np.require(ds[x].values, dtype=np.double, requirements="C")
        for x in ("Longitude_t", "Longitude_u", "Latitude_t", "Latitude_v")
    )

    # Lateral distances
    g["DXGvec"] = g["rSphere"] * np.cos(g["YGvec"] * d2r) / g["resx"] * d2r
    g["DYGsc"] = g["rSphere"] * d2r / g["resy"]
    g["DXCvec"] = g["rSphere"] * np.cos(g["YCvec"] * d2r) / g["resx"] * d2r
    g["DYCsc"] = g["DYGsc"]

    # Vertical coordinate and distances
    g["RC"] = np.require(-ds.Depth_c, dtype=np.double, requirements="C")
    g["DRC"] = np.diff(-g["RC"])

    g["nx"] = g["XCvec"].size
    g["ny"] = g["YCvec"].size
    g["nz"] = g["RC"].size

    # Vertical area of the tracer cells [m^2]
    g["RACvec"] = (g["rSphere"] ** 2 / g["resx"] * d2r) * abs(
        np.sin((g["YGvec"] + 1 / g["resy"]) * d2r) - np.sin(g["YGvec"] * d2r)
    )

    T = ds.theta.isel(Time=ts)
    ds.close()

    ds = xr.open_dataset(file_salt).load()
    S = ds.salt.isel(Time=ts)
    ds.close()

    # Reorder dimensions to ensure individual water columns are float64 and contiguous in memory
    dims = ("Longitude_t", "Latitude_t", "Depth_c")
    S, T = (x.transpose(*dims).astype(np.double, order="C") for x in (S, T))

    return g, S, T



# In[Imports]

import numpy as np

from neutralocean.surface import potential_surf, anomaly_surf, omega_surf
from neutralocean.eos import make_eos_s_t
from neutralocean.ntp import ntp_epsilon_errors

# In[Create an ocean of 4 water columns]

# Make up simple Salt, Temperature, and Pressure data for 4 casts
nc = 4  # Number of casts (water columns)
nk = 20  # Number of grid points per cast
pbot = 4000.0  # Pressure at bottom grid point

P = np.linspace(0, 1, nk) ** 3 * pbot  # pressure increasing cubicly going down

S = np.linspace(34, 36, nk).reshape((1, -1))  # saltier going down
S = S + np.linspace(0, 0.9, nc).reshape((-1, 1))  # some lateral structure

T = np.linspace(14, -2, nk).reshape((1, -1))  # warmer going down
T = T + np.linspace(0, 6, nc).reshape((-1, 1))  # some lateral structure

# Arrange the 4 casts (labelled 0, 1, 2, 3) with connections as follows:
# 0
# | \
# 1--2--3
# That is, cast 0 is connected to casts 1 and 2; cast 1 is connected to casts 0
# and 2; cast 2 is conncted to all casts; cast 3 is connected only to cast 2.
a = np.array([0, 0, 1, 2])  # a[i] and b[i] are a pair of adjacent casts
b = np.array([1, 2, 2, 3])
edges = (a, b)

# Invent distances `dist` between pairs of casts, roughly 100km
dist = np.array([1, 1.4, 1, 1]) * 1e5  # Units: [m]

# Invent distances `distperp` of the interfaces between pairs of casts.
# The product of `dist` and `distperp` gives an area associated to the region
# between casts, which is where the ϵ neutrality errors live. We seek to minimize
# these ϵ neutrality errors, weighted by these areas.
distperp = np.array([1, 1, 1, 1]) * 1e5  # Units: [m]

# Package the grid information into a dict, for neutralocean.
grid = {"edges": edges, "dist": dist, "distperp": distperp}

"""
# Note, more complex examples may have the water column adjacency specified by
# a graph structure, represented as a (sparse) matrix.  For example (with a
# dense matrix here given the smallness of this example), suppose we have two
# graphs, each with the same sparsity structure, one storing the distances
# between adjacent water columns and the other storing the length of the
# interfaces between adjacent water columns.
graph_dist = np.zeros((nc, nc))
graph_dist[0, :] = [0.0, 1.0, 1.4, 0.0]
graph_dist[1, :] = [1.0, 0.0, 1.0, 0.0]
graph_dist[2, :] = [1.4, 1.0, 0.0, 2.0]
graph_dist[3, :] = [0.0, 0.0, 2.0, 0.0]
graph_dist *= 1e5
graph_distperp = np.sign(graph_dist) * 1e5

# To convert these graphs into the format for the `grid` argument to
# neutralocean functions, simply call:
from neutralocean.grid.graph import build_grid
grid = build_grid({"dist": graph_dist, "distperp": graph_distperp})
"""


# In[Approx Neutral Surfaces]

# Here, ν = 1/ρ is the TEOS-10 specific volume,
#       S is Absolute Salinity,
#       Θ is Conservative Temperature,
#       S is Absolute Salinity,
#       p is pressure

# Potential specific volume surface, with given reference pressure and given isovalue.
# This finds the surface satisfying
#   ν(S, Θ, 0 dbar) = (1/1027.5) m3 / kg
s, t, p, d = potential_surf(
    S,
    T,
    P,
    grid=grid,
    eos="gsw",
    ref=0.0,
    isoval=1 / 1027.5,
)
print(
    f" ** The potential specific volume surface (referenced to {d['ref']}dbar)"
    f" with isovalue = {d['isoval']} m3 kg-1"
    f" has root-mean-square ϵ neutrality error {d['e_RMS']} m2 kg-1."
)

# In-situ specific volume anomaly, with given reference S and Θ values and given isovalue.
# This finds the surface satisfying
#   ν(S, Θ, p) - ν(34.5 g/kg, 4.0°C, p) = 0 m3 / kg
s0, t0 = 34.5, 4.0
s, t, p, d = anomaly_surf(
    S,
    T,
    P,
    grid=grid,
    eos="gsw",
    ref=(s0, t0),
    isoval=0.0,
)
print(
    f" ** The in-situ specific volume anomaly surface (referenced to {d['ref']})"
    f" with isovalue = {d['isoval']} m3 kg-1"
    f" has root-mean-square ϵ neutrality error {d['e_RMS']} m2 kg-1."
)


# omega-surface, initialized by a potential density surface, pinning the surface
# to be 1500dbar on cast 0.
s, t, p, d = omega_surf(S, T, P, grid, pin_cast=0, pin_p=1500.0, eos="gsw")
print(
    f" ** The omega-surface"
    f" initialized from a potential density surface (referenced to 1500 dbar)"
    f" intersecting the cast labelled '0' at pressure 1500 bar"
    f" has root-mean-square ϵ neutrality error {d['e_RMS'][-1]} m2 kg-1."
)

# Calculate ϵ neutrality errors on the latest surface, between all pairs of adjacent water columns
eos_s_t = make_eos_s_t("gsw")
e = ntp_epsilon_errors(s, t, p, grid, eos_s_t)
print("The ϵ neutrality errors on the ω-surface are as follows:")
for i in range(len(a)):
    print(f"  From cast {a[i]} to cast {b[i]}, ϵ = {e[i]} m2 kg-1")
print(
    "Note that the connection between casts 2 and 3 has virtually 0 neutrality "
    "error.  This is because cast 3 is ONLY connected to cast 2, so this link "
    "can be along the (discrete) neutral tangent plane joining cast 2 and 3. "
    "The ω-surface finds this."
)



# Show basic use of neutralocean on a tripolar grid using CMIP6 archived
# CanESM5 data, which runs NEMO 3.4 on the ORCA1 grid.
#
# For more advanced usage of neutralocean, see the run_OCCA.py example.

# In[Imports]

import xarray as xr
import pooch

from neutralocean.grid.tripolar import build_grid, edgedata_to_maps
from neutralocean.surface import potential_surf, omega_surf

# In[Load data]

# Get salt and potential temperature from CMIP6 archives (734 MB and 963 MB)
url_salt = "http://crd-esgf-drc.ec.gc.ca/thredds/fileServer/esgC_dataroot/AR6/CMIP6/CMIP/CCCma/CanESM5/esm-piControl/r1i1p1f1/Omon/so/gn/v20190429/so_Omon_CanESM5_esm-piControl_r1i1p1f1_gn_629101-630012.nc"
url_theta = "http://crd-esgf-drc.ec.gc.ca/thredds/fileServer/esgC_dataroot/AR6/CMIP6/CMIP/CCCma/CanESM5/esm-piControl/r1i1p1f1/Omon/thetao/gn/v20190429/thetao_Omon_CanESM5_esm-piControl_r1i1p1f1_gn_629101-630012.nc"

# Get ORCA1 horizontal grid distances from a Zenodo archive that is
# unassociated with the CMIP6 CanESM5 run, I suspect is the same as that which
# would have been used by CanESM5. This example is for illustrative purposes
# only! (405 MB)
url_mesh = "https://zenodo.org/record/4432892/files/ORCA1_mesh_mask.nc"

# Download data, using friendly pooch to fetch it if not already downloaded.
hash_salt = "93ba818ac7661c48686ae3af497e832667d0e72eb1968f375d9191c24f11db94"
hash_theta = "5d4085ab355c3a5526fa62d5f545a2322050e08dc24383189d458b413db99a79"
hash_mesh = "3e4560590c338dfd18de5db580f2354b11acb8ae95e13101c66461683385e777"
file_salt = pooch.retrieve(url=url_salt, known_hash=hash_salt)
file_theta = pooch.retrieve(url=url_theta, known_hash=hash_theta)
file_mesh = pooch.retrieve(url=url_mesh, known_hash=hash_mesh)


# Extract data
timestep = 0
ds = xr.open_dataset(file_salt)
nj = ds.dims["j"]  # number of grid points in the meridional y direction
ni = ds.dims["i"]  # number of grid points in the zonal x direction
S = ds["so"].isel({"time": timestep})  # 3D salinity
Z = ds["lev"]  # 1D depth at center of tracer cells

ds = xr.open_dataset(file_theta)
T = ds["thetao"].isel({"time": timestep})  # potential temperature

mesh = xr.open_dataset(file_mesh)  # contains horizontal grid information

# Squeeze out singleton dimension in horizontal grid distances
e1u, e2v, e2u, e1v = (mesh[x].squeeze() for x in ("e1u", "e2v", "e2u", "e1v"))

# Note: The ORCA1 tripolar horizontal grid is of size (nj, ni) == (291, 360).
# The second dimension is periodic, handling longitude's periodic nature.
# The first dimension has two different boundary conditions. The south
# (the first row) is non-periodic since Antarctica covers the South Pole and
# the ocean grid only goes to -78.3935°S. The north (the last row) is periodic
# with a flipped version of itself, due to ORCA's tripolar grid. Specifically,
# the cell at [nj - 1, i] is adjacent to the cell at [nj - 1, ni - i - 1].
# The grid metrics e1u, e2v, e2u, e1v all have size (nj+1, ni+2) == (292, 362):
# they employ padding to ease applying boundary conditions.
# See documentation in `neutralocean.grid.tripolar` for more information.

# Build the edges of the grid graph, and the metrics associated with each edge
grid = build_grid((nj, ni), e1u, e2v, e2u, e1v)

# In[Approximately Neutral Surfaces]

grav = 9.80665  # gravitational acceleration [m s-2]
rau0 = 1035.0  # Boussinesq reference density [kg m-3]

# Provide reference pressure (actually depth, in Boussinesq) and isovalue
s, t, z, d = potential_surf(
    S,
    T,
    Z,
    grid=grid,
    eos="jmd95",
    grav=grav,
    rho_c=rau0,
    vert_dim="lev",
    ref=0.0,
    isoval=1027.5,
)
s_sigma, t_sigma, z_sigma = s, t, z  # save for later
print(
    f" ** The potential density surface (referenced to {d['ref']}m)"
    f" with isovalue = {d['isoval']}kg m-3"
    f" has root-mean-square ϵ neutrality error {d['e_RMS']} kg m-4"
)

# Pin omega surface to the cast (j0, i0) in the equatorial Pacific at depth 1500m.
j0, i0 = 100, 150
z0 = 1500.0

# Initialize omega surface with a (locally referenced) potential density surface.
# Provide grid distances. By default, does as many iterations as needed to get
# the root-mean-square change in the locally referenced potential density from
# one iteration to the next to drop below 10^-7 kg m-3, or a maximum of 10
# iterations.
s, t, z, d = omega_surf(
    S,
    T,
    Z,
    grid,
    vert_dim="lev",
    pin_cast=(j0, i0),
    pin_p=z0,
    eos="jmd95",
    grav=grav,
    rho_c=rau0,
)
s_omega, t_omega, z_omega = s, t, z  # save for later
print(
    f" ** The omega-surface"
    f" initialized from a potential density surface (referenced to {z0}m)"
    f" intersecting the cast indexed by {(j0,i0)} at depth {z0}m"
    f" has root-mean-square ϵ neutrality error {d['e_RMS'][-1]} kg m-4"
)


# In[Neutrality errors on a surface]
from neutralocean.ntp import ntp_epsilon_errors
from neutralocean.eos import make_eos_s_t

# Prepare function for S and T derivatives of the equation of state
eos_s_t = make_eos_s_t("jmd95", grav, rau0)

# Calculate ϵ neutrality errors on all pairs of adjacent water columns
e = ntp_epsilon_errors(s, t, z, grid, eos_s_t)

# Convert ϵ above into two 2D maps, one each of the "x" and "y" dimensions
ex, ey = edgedata_to_maps(e, (nj, ni))

# These can then be mapped:
# import matplotlib.pyplot as plt
# plt.imshow(ex, origin="lower", vmin=-1e-9, vmax=1e-9)
# plt.colorbar()



import numpy as np
import numba as nb

from .lib import diff_1d_samesize, valid_range_1_two

from .ppinterp import ppinterp_1, ppinterp_1_two


def pchip_coeffs(X, Y):
    """
    Piecewise Polynomial Coefficients for a PCHIP
    (Piecewise Cubic Hermite Interpolating Polynomial)

    Parameters
    ----------
    X : ndarray

        Independent variable. Must be same dimensions as `Y`, with any
        dimension possibly a singleton: that is, must be broadcastable
        to the size of `Y`.

        `X` must be monotonically increasing in its last dimension.
        That is, `X[*i,:]` should be monotonically increasing for any
        `i` tuple indexing all but the last dimension.

    Y : ndarray

        Dependent variable.  Must be same dimensions as `X`, with any
        dimension possibly a singleton: that is, must be broadcastable
        to the size of `X`.

    Returns
    -------
    Yppc : ndarray

        Coefficients for a PCHIP interpolant of `Y` to `X`.

    Notes
    -----
    If `X` and `Y` have NaN's, only the first contiguous block of non-NaN data
    between both `X` and `Y` is used.

    Evaluate the piecewise polynomial at `x` as

    >>> y = ppval(x, X, Yppc)

    """

    # Suppress the divide by zero and invalid arguments warnings that appear
    # for the `numba.guvectorize`d function, though not for the `numba.njit`ed
    # function.  The PCHIP algorithm is guaranteed to not divide by zero.
    # https://github.com/numba/numba/issues/4793#issuecomment-623296850
    with np.errstate(divide="ignore", invalid="ignore"):
        return _pchip_coeffs(X, Y, np.arange(4))


def pchip_interp(x, X, Y, d=0):

    # Suppress the divide by zero and invalid arguments warnings that appear
    # for the `numba.guvectorize`d function, though not for the `numba.njit`ed
    # function.  The PCHIP algorithm is guaranteed to not divide by zero.
    # https://github.com/numba/numba/issues/4793#issuecomment-623296850
    with np.errstate(divide="ignore", invalid="ignore"):
        return _pchip_interp(x, X, Y, d)


def pchip_interp_two(x, X, Y, Z, d=0):

    # Suppress the divide by zero and invalid arguments warnings that appear
    # for the `numba.guvectorize`d function, though not for the `numba.njit`ed
    # function.  The PCHIP algorithm is guaranteed to not divide by zero.
    # https://github.com/numba/numba/issues/4793#issuecomment-623296850
    with np.errstate(divide="ignore", invalid="ignore"):
        return _pchip_interp_two(x, X, Y, Z, d)


@nb.njit
def pchip_coeffs_1(X, Y):
    """
    Coefficients for a single PCHIP when the data may contain NaNs

    Inputs and outputs are as for `pchip_coeffs`, but `X` and `Y` both 1D
    arrays of length `n`, and so `Yppc` is a 2D array of size `(n, 4)`.
    """

    # Find k = index to first valid data site, and
    # K such that K - 1 = index to last valid data site in contiguous range
    #                     of valid data after index k.
    k, K = valid_range_1_two(X, Y)

    return _pchip_coeffs_1(X, Y, k, K)


@nb.njit
def pchip_coeffs_1_nonan(X, Y):
    """
    Coefficients for a single PCHIP when the data has no NaNs

    Inputs and outputs are as for `pchip_coeffs`, but `X` and `Y` both 1D
    arrays of length `n`, and so `Yppc` is a 2D array of size `(n, 4)`.
    """

    return _pchip_coeffs_1(X, Y, 0, X.size)


@nb.njit
def _pchip_coeffs_1(X, Y, k, K):
    """
    Coeffs for a single PCHIP when data is non-NaN on `k:K`
    """

    # Note, np.diff does not work with nb.njit, here.
    # Also, using our own numba.njit'ed function is faster than np.ediff1d
    h = diff_1d_samesize(X)  # distance between data sites
    δ = diff_1d_samesize(Y) / h  # linear slope between data sites

    nk = X.size  # Number of data points included possible NaN's

    d = np.zeros(K)  # slope of interpolant at data sites
    #                  Note elements 0, ... k-1 of d will be unused.
    C = np.full((nk, 4), np.nan)

    if K - k > 2:

        # Calculate PCHIP slopes
        #  Slopes at end points:
        #   Set d[k] and d[K-1] via non-centered, shape-preserving three-point formulae.
        #  Slopes at interior points:
        #   d[i] = weighted average of δ[i-1] and δ[i] when they have the same sign.
        #   d[i] = 0 when δ[i-1] and δ[i] have opposites signs or either is zero.

        d[k] = ((2 * h[k] + h[k + 1]) * δ[k] - h[k] * δ[k + 1]) / (
            h[k] + h[k + 1]
        )
        if np.sign(d[k]) != np.sign(δ[k]):
            d[k] = 0
        elif (np.sign(δ[k]) != np.sign(δ[k + 1])) and (
            abs(d[k]) > abs(3 * δ[k])
        ):
            d[k] = 3 * δ[k]

        for i in range(k + 1, K - 1):
            if np.sign(δ[i - 1]) * np.sign(δ[i]) > 0:
                w1 = h[i - 1] + 2 * h[i]
                w2 = 2 * h[i - 1] + h[i]
                d[i] = (w1 + w2) / (w1 / δ[i - 1] + w2 / δ[i])
            else:
                d[i] = 0

        d[K - 1] = (
            (2 * h[K - 2] + h[K - 3]) * δ[K - 2] - h[K - 2] * δ[K - 3]
        ) / (h[K - 2] + h[K - 3])
        if np.sign(d[K - 1]) != np.sign(δ[K - 2]):
            d[K - 1] = 0
        elif (np.sign(δ[K - 2]) != np.sign(δ[K - 3])) and (
            abs(d[K - 1]) > abs(3 * δ[K - 2])
        ):
            d[K - 1] = 3 * δ[K - 2]

        # Build piecewise cubic Hermite polynomial
        for i in range(k, K - 1):
            dzzdx = (δ[i] - d[i]) / h[i]
            dzdxdx = (d[i + 1] - δ[i]) / h[i]
            C[i, 0] = (dzdxdx - dzzdx) / h[i]
            C[i, 1] = 2 * dzzdx - dzdxdx
            C[i, 2] = d[i]

    elif K - k == 2:
        # Special case: use linear interpolation.

        δ = (Y[k + 1] - Y[k]) / (X[k + 1] - X[k])

        C[k, 0] = 0
        C[k, 1] = 0
        C[k, 2] = δ

    # else:  # K - k == 1
    #     leave coefficients as nans.  Ignore case of x = X[k] while
    #     having X[k+1:] == np.nan and/or Y[k+1:] == np.nan.  That case
    #     cannot be handled here, since setting the coefficients to
    #     be [0, 0, 0, Y[k]], i.e. a constant function, would mean the
    #     interpolant would be Y[k] when evaluated 0at x > X[k], but we
    #     only want it to be Y[k] at x = X[k] precisely.

    C[:, 3] = Y

    return C


@nb.njit
def _pchip_coeffs_i(X, Y, i):
    """
    Coefficients of a single Piece of a PCHIP

    Parameters
    ----------
    X : 1D array
        Independent data, monotonically increasing

    Y : 1D array
        Dependent data

    i : int
        Select the interval of `X` that contains the eventual evaluation site
        `x`. Specifically,
            if `i == 0` then `X[0] <= x <= X[1]`, or
            if `1 <= i <= len(X) - 1` then `X[i] < x <= X[i+1]`.

        These facts about `i` are assumed true; they are not checked.

    Returns
    -------
    C3, C2, C1, C0 : float
        Piecewise Polynomial Coefficients that can be evaluated at `x`.

    Notes
    -----
    To interpolate `Y` in terms of `X` at evaluation site `x`, simply evaluate
    the piecewise polynomial whose coefficients are `Yppc` at `x` by
        `pval(x - X[i], (C3, C2, C1, C0))`

    This function is adapted from `_pchip_coeffs_1`
    """

    # Pre-assign sizes for PCHIP variables.
    h = [0.0, 0.0, 0.0]
    δ = [0.0, 0.0, 0.0]
    d = [0.0, 0.0]

    # Check whether x is adjacent to the start or end of this X
    at_start = (i == 0) or np.isnan(X[i - 1] + Y[i - 1])
    at_end = (i == len(X) - 2) or np.isnan(X[i + 2] + Y[i + 2])

    if at_start and at_end:

        # if np.isnan(X[i + 1]) or np.isnan(Y[i + 1]):
        #     # Only one valid data point.  Leave the interpolant as NaN.
        #     d[0], c, b = np.nan, np.nan, np.nan

        # else:

        # ||| X[0] <= x <= X[1] |||   Revert to Linear Interpolation
        # If actually only one non-NaN data point, then d[0] will be NaN, so
        # interpolant will evaluate to NaN.
        d[0] = (Y[i + 1] - Y[i]) / (X[i + 1] - X[i])
        C3, C2 = 0.0, 0.0

    else:
        if at_start:
            #  ||| X[0] <= x <= X[1] < X[2] --->
            h[1] = X[i + 1] - X[i]
            h[2] = X[i + 2] - X[i + 1]
            δ[1] = (Y[i + 1] - Y[i]) / h[1]
            δ[2] = (Y[i + 2] - Y[i + 1]) / h[2]

            #  Noncentered, shape-preserving, three-point formula:
            d[0] = ((2.0 * h[1] + h[2]) * δ[1] - h[1] * δ[2]) / (h[1] + h[2])
            if np.sign(d[0]) != np.sign(δ[1]):
                d[0] = 0.0
            elif (np.sign(δ[1]) != np.sign(δ[2])) and (
                np.abs(d[0]) > np.abs(3.0 * δ[1])
            ):
                d[0] = 3.0 * δ[1]

            # Standard PCHIP formula
            if np.sign(δ[1]) * np.sign(δ[2]) > 0.0:
                w1 = 2.0 * h[2] + h[1]
                w2 = h[2] + 2.0 * h[1]
                d[1] = (w1 + w2) / (w1 / δ[1] + w2 / δ[2])
            else:
                d[1] = 0.0

        elif at_end:
            # <--- X[i-1] < X[i] < x <= X[i+1] |||
            h[0] = X[i] - X[i - 1]
            h[1] = X[i + 1] - X[i]
            δ[0] = (Y[i] - Y[i - 1]) / h[0]
            δ[1] = (Y[i + 1] - Y[i]) / h[1]

            # Standard PCHIP formula
            if np.sign(δ[0]) * np.sign(δ[1]) > 0.0:
                w1 = 2.0 * h[1] + h[0]
                w2 = h[1] + 2.0 * h[0]
                d[0] = (w1 + w2) / (w1 / δ[0] + w2 / δ[1])
            else:
                d[0] = 0.0

            #  Noncentered, shape-preserving, three-point formula:
            d[1] = ((h[0] + 2.0 * h[1]) * δ[1] - h[1] * δ[0]) / (h[0] + h[1])
            if np.sign(d[1]) != np.sign(δ[1]):
                d[1] = 0.0
            elif (np.sign(δ[1]) != np.sign(δ[0])) and (
                np.abs(d[1]) > np.abs(3 * δ[1])
            ):

                d[1] = 3.0 * δ[1]

        else:
            # <--- X[i-1] < X[i] < x <= X[i+1] < X[i+2] --->
            h[0] = X[i] - X[i - 1]  # Way faster to do this
            h[1] = X[i + 1] - X[i]  # than
            h[2] = X[i + 2] - X[i + 1]  # diff(X(i-1:i+3))
            δ[0] = (Y[i] - Y[i - 1]) / h[0]
            δ[1] = (Y[i + 1] - Y[i]) / h[1]
            δ[2] = (Y[i + 2] - Y[i + 1]) / h[2]

            # Standard PCHIP formula
            for j in range(2):
                if np.sign(δ[j]) * np.sign(δ[j + 1]) > 0.0:
                    w1 = 2.0 * h[j + 1] + h[j]
                    w2 = h[j + 1] + 2.0 * h[j]
                    d[j] = (w1 + w2) / (w1 / δ[j] + w2 / δ[j + 1])
                else:
                    d[j] = 0.0

        # Polynomial coefficients for this piece
        dzzdx = (δ[1] - d[0]) / h[1]
        dzdxdx = (d[1] - δ[1]) / h[1]
        C3 = (dzdxdx - dzzdx) / h[1]  # coeff of the 3rd degree term (x^3)
        C2 = 2 * dzzdx - dzdxdx  # coeff of 2nd degree term (x^2)

    # The following code evaluates the `d`'th deriviative of the cubic
    # interpolant at `x`.
    # s = x - X[i]
    # if d == 0:
    #     y = Y[i] + s * (d[0] + s * (C2 + s * C3))
    # elif d == 1:  # first derivative
    #     y = d[0] + s * (2 * C2 + 3 * s * C3)
    # elif d == 2:  # second derivative
    #     y = 2 * C2 + 6 * s * C3
    # elif d == 3:  # third derivative
    #     y = 6 * C3
    # else:
    #     y = 0.0
    # return y

    # Faster to return tuple than build an np.array just to deconstruct it later
    return C3, C2, d[0], Y[i]


@nb.guvectorize(
    [(nb.f8[:], nb.f8[:], nb.f8[:], nb.f8[:, :])], "(n),(n),(m)->(n,m)"
)
def _pchip_coeffs(X, Y, len4array, Yppc):
    Yppc[:, :] = pchip_coeffs_1(X, Y)


@nb.njit
def pchip_interp_1(x, X, Y, d=0):
    return ppinterp_1(_pchip_coeffs_i, x, X, Y, d)


@nb.njit
def pchip_interp_1_two(x, X, Y, Z, d=0):
    return ppinterp_1_two(_pchip_coeffs_i, x, X, Y, Z, d)


@nb.guvectorize(
    [(nb.f8, nb.f8[:], nb.f8[:], nb.i8, nb.f8[:])],
    "(),(n),(n),()->()",
)
def _pchip_interp(x, X, Y, d, y):
    y[0] = pchip_interp_1(x, X, Y, d)


@nb.guvectorize(
    [(nb.f8, nb.f8[:], nb.f8[:], nb.f8[:], nb.i8, nb.f8[:], nb.f8[:])],
    "(),(n),(n),(n),()->(),()",
)
def _pchip_interp_two(x, X, Y, Z, d, y, z):
    y[0], z[0] = pchip_interp_1_two(x, X, Y, Z, d)



"""Neutral Tangent Plane and related functions"""

import numpy as np
import numba as nb

from neutralocean.lib import xr_to_np
from neutralocean.eos.tools import make_eos_s_t


def ntp_epsilon_errors(s, t, p, grid, eos_s_t="gsw", grav=None, rho_c=None):
    """
    Calculate epsilon neutrality errors on an approximately neutral surface

    Parameters
    ----------
    s, t, p, eos_s_t, grav, rho_c :
        See `ntp_epsilon_errors_norms`.

    grid : dict
        See `ntp_epsilon_errors_norms`.
        Need not have the `distperp` element, as this is not used.
        If the `dist` element is missing, a value of 1.0 will be used.
        Can alternatively pass a 2 element tuple that is just `grid['edges']`,
        in which case `dist` will be taken as 1.0.


    Returns
    -------
    e : array
        The epsilon neutrality errors on the surface.
        `e[i]` is the neutrality error from node `a[i]` to node `b[i]`, where
        `a, b = grid['edges']`.
    """

    if isinstance(grid, tuple):
        edges = grid
        dist = 1.0
    elif isinstance(grid, dict):
        edges = grid["edges"]
        dist = grid.get("dist", 1.0)

    eos_s_t = make_eos_s_t(eos_s_t, grav, rho_c)

    s, t, p = (np.reshape(xr_to_np(x), -1) for x in (s, t, p))
    e = _ntp_epsilon_error1(s, t, p, edges[0], edges[1], eos_s_t)

    if dist is not float or dist != 1.0:
        e = e / dist
    return e


def ntp_epsilon_errors_norms(
    s, t, p, grid, eos_s_t="gsw", grav=None, rho_c=None
):
    """
    Calculate norms of the epsilon neutrality errors on an approximately neutral surface

    Parameters
    ----------
    s, t, p : ndarray
        practical / Absolute Salinity, potential / Conservative temperature,
        and pressure / depth on the surface

    grid : dict
        Containing the following:

        edges : tuple of length 2, Required
            Each element is an array of int of length E, where E is the number of
            edges in the grid's graph, i.e. the number of pairs of adjacent water
            columns (including land) in the grid.
            If `edges = (a, b)`, the nodes (water columns) whose linear indices are
            `a[i]` and `b[i]` are adjacent.

        dist : 1d array, Default 1.0
            Horizontal distance between adjacent water columns (nodes).
            `dist[i]` is the distance between nodes whose linear indices are
            `edges[0][i]` and `edges[1][i]`.

        distperp : 1d array, Default 1.0
            Horizontal distance of the face between adjacent water columns (nodes).
            `distperp[i]` is the distance of the interface between nodes whose
            linear indices are `edges[0][i]` and `edges[1][i]`.

    eos_s_t : function
        Equation of state for the partial derivatives of density or specific
        volume with respect to `S` and `T` as a function of `S`, `T`, and `P`
        inputs.

    grav : float, Default None
        Gravitational acceleration [m s-2].  When non-Boussinesq, pass `None`.

    rho_c : float, Default None
        Boussinesq reference density [kg m-3].  When non-Boussinesq, pass `None`.

    Returns
    -------
    e_RMS : float
        Area-weighted root-mean-square of the epsilon neutrality error on the surface

    e_MAV : TYPE
        Area-weighted mean-absolute-value of the epsilon neutrality error on the surface

    """

    eos_s_t = make_eos_s_t(eos_s_t, grav, rho_c)

    # Calculate epsilon neutrality errors.  Here, treat all distances = 1.
    # The actual distances will be handled in computing the norms
    e = ntp_epsilon_errors(s, t, p, {"edges": grid["edges"]}, eos_s_t)

    area = grid["dist"] * grid["distperp"]  # Area [m^2] centred on edges

    # L2 norm of vector [a_i], weighted by vector [w_i], is sqrt( sum( w_i * a_i^2 ) / sum( w_i ) )
    # Here, weights are `area`.
    # But also need to divide epsilon by grid distances `dist`.
    # Thus, the numerator of L2 norm needs to multiply epsilon^2 by
    #     area / dist^2 = distperp / dist,
    e_RMS = np.sqrt(
        np.nansum(grid["distperp"] / grid["dist"] * e**2)
        / np.sum(area * np.isfinite(e))
    )

    # L1 norm of vector [a_i], weighted by vector [w_i], is sum( w_i * |a_i| ) / sum( w_i )
    # Here, weights are `area`.
    # But also need to divide epsilon by grid distances `dist`.
    # Thus, the numerator of L1 norm needs to multiply epsilon by
    #     area ./ dist = distperp ,
    e_MAV = np.nansum(grid["distperp"] * abs(e)) / np.sum(
        area * np.isfinite(e)
    )

    return e_RMS, e_MAV


@nb.njit
def avg1(a, b):
    return (a + b) * 0.5


@nb.njit
def dif1(a, b):
    return b - a


@nb.njit
def arg1(a, b):
    return a


@nb.njit
def arg2(a, b):
    return b


@nb.njit
def _ntp_epsilon_error1(s, t, p, a, b, eos_s_t):
    e = np.empty(len(a), dtype=s.dtype)
    for i in range(len(a)):
        a_ = a[i]
        b_ = b[i]
        rs, rt = eos_s_t(
            0.5 * (s[a_] + s[b_]),
            0.5 * (t[a_] + t[b_]),
            0.5 * (p[a_] + p[b_]),
        )
        e[i] = rs * (s[b_] - s[a_]) + rt * (t[b_] - t[a_])
    return e



"""Library of simple functions for neutral_surfaces"""

import numpy as np
import numba as nb
import xarray as xr

from .eos import make_eos, make_eos_s_t


@nb.njit
def find_first_nan(a):
    """The index to the first NaN along the last axis

    Parameters
    ----------
    a : ndarray
        Input array possibly containing some NaN elements

    Returns
    -------
    k : ndarray of int
        The index to the first NaN along each 1D array making up `a`, as in the
        following example with `a` being 3D.
        If all `a[i,j,:]` are NaN, then `k[i,j] = 0`.
        If all `a[i,j,:]` are not NaN, then `k[i,j] = a.shape[-1]`.
        Otherwise, `K = k[i,j]` is the smallest int such that `a[i,j,K-1]`
        is not NaN, but `a[i,j,K]` is NaN.
    """
    nk = a.shape[-1]
    k = np.full(a.shape[:-1], nk, dtype=np.int_)
    for n in np.ndindex(a.shape[0:-1]):
        for i in range(nk):
            if np.isnan(a[n][i]):
                k[n] = i
                break
    return k


@nb.njit
def take_fill(a, idx, fillval=np.nan):
    """
    Like numpy.take but fills with nan when indices are out of range

    Parameters
    ----------
    a : ndarray
        input data

    idx : 1d array
        linear indices to elements of `a`

    Returns
    -------
    b : ndarray
        The i'th element of b (in linear order) is the `idx[i]`'th element of `a`,
        (in linear order), or nan if `idx[i] < 0`.  Same shape as `idx`.
    """
    b = np.empty(idx.size, dtype=a.dtype)
    a_ = a.reshape(-1)
    for i in range(len(b)):
        if idx[i] >= 0:
            b[i] = a_[idx[i]]
        else:
            b[i] = fillval
    return b


@nb.njit
def aggsum(a, idx, n):
    """
    Aggregate data into groups and then sum each group.

    Parameters
    ----------
    a : array
        Input data to be aggregated into groups and summed.

    idx : array of int
        Group label for each element of `a`.  To exclude element `i` of `a`
        from any group, let `idx[i]` be a negative int.  Must be same size
        as `a`.

    n : int
        Number of groups, including empty groups.
        As this is also the length of `b`, must satisfy `n >= np.max(idx) + 1`.

    Returns
    -------
    b : array
        The sum of each group of data from `a`.

    Notes
    -----
    This is a simple implementation of `numpy_groupies.aggregate`.
    See https://github.com/ml31415/numpy-groupies/

    """
    b = np.zeros(n, dtype=a.dtype)
    for i in range(len(idx)):
        if idx[i] >= 0:
            b[idx[i]] += a[i]
    return b


def val_at(T, k):
    """Evaluate nD array at given indices along its last dimension

    Parameters
    ----------
    T : ndarray
        Input array. Can be 1D or nD.

    k : int or ndarray
        Index at which to evaluate `T` along its last dimension.
        Can be an int or (n-1)D.

    Returns
    -------
    Tk : ndarray
        The input `T` evaluated with its last index equal to `k`.


    Notes
    -----
    If `T` is 3D and `k` is 2D, then `Tk[i,j] = T[i,j,k[i,j]]` for
    each valid `(i,j)`.

    If `T` is 1D and `k` is 2D, then `Tk[i,j] = T[k[i,j]]` for
    each valid `(i,j)`.

    If `T` is 3D and `k` is an int, then `Tk[i,j] = T[i,j,k]` for
    each valid `(i,j)`.

    Examples
    --------
    Evaluate temperature, having data in each water column, at the bottom grid cell

    >>> T = np.empty((3, 2, 10))  # (longitude, latitude, depth), let us say
    >>> T[..., :] = np.arange(10, 0, -1)  # decreasing along depth dim from 10 to 1
    >>> T[0, 0, :] = np.nan  # make cast (0,0) be land
    >>> T[0, 1, 3:] = np.nan  # make cast (0,1) be only 3 ocean cells deep
    >>> n_good = find_first_nan(T)
    >>> val_at(T, n_good - 1)
    array([[nan,  8.], [ 1.,  1.], [ 1.,  1.]])

    Evaluate the depth at the bottom grid cell

    >>> Z = np.linspace(0, 4500, 10)  # grid cell centre's are at depths 0, 500, 1000, ..., 4500.
    >>> val_at(Z, n_good - 1)  # Z doesn't have NaN structure, so use n_good from T as above
    array([[  nan, 1000.], [4500., 4500.], [4500., 4500.]])
    """
    if isinstance(k, int) or T.ndim == 1:
        # select the k'th element along the last dimension of T
        Tk = T[..., k]
    elif T.ndim == k.ndim + 1:
        # if k[i,j] == 0, this will index T[i,j,-1] which will be nan, so T_bot[i,j] == nan.
        Tk = np.take_along_axis(T, k[..., None], -1).squeeze()
    else:
        raise ValueError(
            "T must be 1 dimensional or have 1 more dimension than k"
        )

    # Set to NaN any place where k is negative
    Tk[k < 0] = np.nan
    return Tk


def xr_to_np(S):
    """Convert xarray into numpy array"""
    if hasattr(S, "values"):
        S = S.values
    return S


def _xr_in(S, vert_dim):
    # Prepare xarray container for output: like input S but without dimension
    # labelled `drop_dim`
    if isinstance(S, xr.core.dataarray.DataArray):
        if vert_dim is None:
            return xr.full_like(S, 0)
        elif isinstance(vert_dim, int):
            vert_dim = S.dims[vert_dim]  # convert to str
        return xr.full_like(S.isel({vert_dim: 0}).drop_vars(vert_dim), 0)
    else:
        return None


def _xrs_in(S, T, P, vert_dim):
    # Prepare xarray containers for output: like inputs S, T, P but without
    # the dimension labelled `vert_dim`.  Doing S, T, P together allows for
    # pxr to be an xarray even if P is an ndarray -- it just won't have attributes.
    sxr, txr = (_xr_in(X, vert_dim) for X in (S, T))

    if sxr is None:
        pxr = None
    else:
        pxr = sxr.copy()
        try:
            pxr.attrs.update(P.attrs)
        except:
            pxr.attrs.clear()

    return sxr, txr, pxr


def _xr_out(s, sxr):
    # Return xarrays if inputs were xarrays
    if isinstance(sxr, xr.core.dataarray.DataArray):
        sxr.data = s
        return sxr
    else:
        return s


def _process_vert_dim(vert_dim, S):
    """Convert `vert_dim` as a str naming a dimension in `S` or a (possibly
    negative) int into an int between 0 and S.ndim-1."""
    if isinstance(vert_dim, str) and hasattr(S, "dims"):
        try:
            vert_dim = S.dims.index(vert_dim)
        except:
            raise ValueError(f"vert_dim = {vert_dim} not found in S.dims")

    return np.mod(vert_dim, S.ndim)


def _contiguous_casts(S, vert_dim=-1):
    """Make individual casts contiguous in memory

    Parameters
    ----------
    S : ndarray
        ocean data such as salinity, temperature, or pressure

    vert_dim : int, Default -1
        Specifies which dimension of `S` is vertical.

    Returns
    -------
    S : ndarray
        input data, possibly re-arranged to have `vert_dim` the last dimension
    """

    if S.ndim > 1 and vert_dim not in (-1, S.ndim - 1):
        S = np.moveaxis(S, vert_dim, -1)

    return np.require(S, dtype=np.float64, requirements="C")


def _process_casts(S, T, P, vert_dim):
    """Make individual casts contiguous in memory and extract numpy array from xarray"""

    vert_dim = _process_vert_dim(vert_dim, S)

    S, T, P = (xr_to_np(x) for x in (S, T, P))

    # Broadcast a 1D vector for P into a ND array like S
    if P.ndim < S.ndim:
        # First make P a 3D array with its non-singleton dimension be `vert_dim`
        P = np.reshape(
            P, tuple(-1 if x == vert_dim else 1 for x in range(S.ndim))
        )
        P = np.broadcast_to(P, S.shape)

    S, T, P = (_contiguous_casts(x, vert_dim) for x in (S, T, P))
    return S, T, P


def _interp_casts(S, T, P, interp_fn, Sppc=None, Tppc=None):
    # Compute interpolants for S and T casts (unless already provided)
    ni, nj, nk = S.shape
    if Sppc is None or Sppc.shape[0:-1] != (ni, nj, nk - 1):
        Sppc = interp_fn(P, S)
    if Tppc is None or Tppc.shape[0:-1] != (ni, nj, nk - 1):
        Tppc = interp_fn(P, T)
    return Sppc, Tppc


def _process_wrap(wrap, s=None, diags=False):
    """Convert to a tuple of `int`s specifying which horizontal dimensions are periodic"""

    if wrap is None:
        if diags:
            raise ValueError(
                "wrap must be given for omega surfaces, or when `diags` is True"
            )
        else:
            return wrap

    if isinstance(wrap, str):
        wrap = (wrap,)  # Convert single string to tuple
    if not isinstance(wrap, (tuple, list)):
        raise TypeError("If given, wrap must be a tuple or list or str")
    if all(isinstance(x, str) for x in wrap):
        try:
            # Convert dim names to tuple of bool
            wrap = tuple(x in wrap for x in s.dims)
        except:
            raise TypeError(
                "With wrap provided as strings, s must have a .dims attribute"
            )

    # type checking on final value
    if not (isinstance(wrap, (tuple, list)) and len(wrap) == 2):
        raise TypeError(
            "wrap must be a two element (logical) array "
            "or a string (or array of strings) referring to dimensions in xarray S"
        )
    return wrap


def _process_pin_cast(pin_cast, S):
    """
    If pinning cast is a dict:
        convert from a coordinate representation,
        suitable for `S.sel(pin_cast)` where S is an xarray,
        into an index representation,
        suitable for `S[pin_cast]` where S is an ndarray.
    If pinning cast is an int:
        wrap it into a 1-element tuple, so np.ravel_multi_index works
    Otherwise, just return the input `pin_cast`.
    """
    # TODO: There must be a better way of doing this...
    # One issue is this always rounds one way, whereas a "nearest" neighbour
    # type behaviour would be preferred, as in xr.DataArray.sel
    if isinstance(pin_cast, dict):
        return tuple(
            int(S.get_index(k).searchsorted(v)) for (k, v) in pin_cast.items()
        )
    elif isinstance(pin_cast, int):
        return (pin_cast,)
    else:
        return pin_cast


def _process_eos(eos, grav=None, rho_c=None, need_s_t=False):
    # Process equation of state argument and make cache functions

    eos_s_t = None
    if isinstance(eos, str):
        if need_s_t:
            eos_s_t = make_eos_s_t(eos, grav, rho_c)
        eos = make_eos(eos, grav, rho_c)
    else:
        if need_s_t:
            if isinstance(eos, (tuple, list)) and len(eos) == 2:
                eos_s_t = eos[1]
                eos = eos[0]
            if not callable(eos) or not callable(eos_s_t):
                raise ValueError(
                    "If `eos` is not a str, expected a tuple of length two"
                    " containing an eos function and an eos_s_t function."
                )
        else:
            if isinstance(eos, (tuple, list)) and len(eos) >= 1:
                eos = eos[0]
            if not callable(eos):
                raise ValueError("If `eos` is not a str, expected a function.")
    return eos, eos_s_t



# In[Imports]

# Functions to make the Equation of State
from neutralocean.eos import make_eos, make_eos_s_t

# Functions to compute various approximately neutral surfaces
from neutralocean.surface import potential_surf, anomaly_surf, omega_surf

# Functions to load OCCA data
from neutralocean.examples.load_OCCA import load_OCCA

from neutralocean.grid.rectilinear import build_grid, edgedata_to_maps

# In[Load OCCA data]

g, S, T = load_OCCA()  # S, T arranged as (Longitude, Latitude, Depth)
ni, nj, nk = S.shape
Z = -g["RC"]  # Depth vector (note positive and increasing down)

# Select pinning cast in the equatorial Pacific. 
# The following surfaces will intersect cast (i0,j0) at a depth of z0.
i0, j0 = 220, 80
z0 = 1500.0

# make Boussinesq version of the Jackett and McDougall (1995) equation of state
# --- which is what OCCA used --- and its partial derivatives
eos = make_eos("jmd95", g["grav"], g["rho_c"])
eos_s_t = make_eos_s_t("jmd95", g["grav"], g["rho_c"])

# Build grid adjacency and distance information for neutralocean functions
grid = build_grid(
    (ni, nj), g["wrap"], g["DXCvec"], g["DYCsc"], g["DYGsc"], g["DXGvec"]
)

# In[Potential Density surfaces]

# Provide reference pressure (actually depth, in Boussinesq) and isovalue
s, t, z, d = potential_surf(
    S,
    T,
    Z,
    grid=grid,
    eos="jmd95",
    grav=g["grav"],
    rho_c=g["rho_c"],
    vert_dim="Depth_c",
    ref=0.0,
    isoval=1027.5,
)
print(
    f" ** The potential density surface (referenced to {d['ref']}m)"
    f" with isovalue = {d['isoval']}kg m-3"
    f" has root-mean-square ϵ neutrality error {d['e_RMS']} kg m-4"
)

# Provide pin_cast and pin_p: the reference location and depth that the surface intersects
s, t, z, d = potential_surf(
    S,
    T,
    Z,
    grid=grid,
    eos="jmd95",
    grav=g["grav"],
    rho_c=g["rho_c"],
    vert_dim="Depth_c",
    ref=0.0,
    pin_cast=(i0, j0),
    pin_p=z0,
)
print(
    f" ** The potential density surface (referenced to {d['ref']}m)"
    f" intersecting the cast indexed by {(i0,j0)} at depth {z0}m"
    f" (isovalue = {d['isoval']}kg m-3)"
    f" has root-mean-square ϵ neutrality error {d['e_RMS']} kg m-4"
)

# Provide just the location to intersect `(pin_cast, pin_p)`.
# This takes the reference depth `ref` to match `pin_p`.
# Also illustrate using xarray coordinates for pin_cast.
# Also use PCHIPs as the vertical interpolants.
# Also provide the (pre-made) equation of state and its partial derivatives w.r.t S and T.
s, t, z, d = potential_surf(
    S,
    T,
    Z,
    grid=grid,
    eos=(eos, eos_s_t),
    vert_dim="Depth_c",
    pin_cast={"Longitude_t": 220.5, "Latitude_t": 0.5},
    pin_p=z0,
    interp="pchip",
)
assert z.values[i0, j0] == z0  # check pin_cast was indeed (i0,j0)
z_sigma = z  # save for later
print(
    f" ** The potential density surface (referenced to {d['ref']}m)"
    f" intersecting the cast at (180.5 E, 0.5 N) at depth {z0}m"
    f" (isovalue = {d['isoval']}kg m-3)"
    f" has root-mean-square ϵ neutrality error {d['e_RMS']} kg m-4"
)

# In[Delta surfaces]

# Provide reference salinity and potential temperature values
s0, t0 = 34.5, 4.0
s, t, z, d = anomaly_surf(
    S,
    T,
    Z,
    grid=grid,
    eos=(eos, eos_s_t),
    vert_dim="Depth_c",
    ref=(s0, t0),
    isoval=0.0,
)
print(
    f" ** The in-situ density anomaly surface (referenced to {d['ref']})"
    f" with isovalue = {d['isoval']}kg m-3"
    f" has root-mean-square ϵ neutrality error {d['e_RMS']} kg m-4"
)

# Provide pin_cast and pin_p: the reference location and depth that the surface intersects
s, t, z, d = anomaly_surf(
    S,
    T,
    Z,
    grid=grid,
    eos=(eos, eos_s_t),
    vert_dim="Depth_c",
    ref=(s0, t0),
    pin_cast=(i0, j0),
    pin_p=z0,
)
print(
    f" ** The in-situ density anomaly surface (referenced to {d['ref']})"
    f" intersecting the cast indexed by {(i0,j0)} at depth {z0}m"
    f" (isovalue = {d['isoval']}kg m-3)"
    f" has root-mean-square ϵ neutrality error {d['e_RMS']} kg m-4"
)

# Provide just the location to intersect: depth `pin_p` on cast `pin_cast`
# This takes the reference S and T values from that location.
s, t, z, d = anomaly_surf(
    S,
    T,
    Z,
    grid=grid,
    eos=(eos, eos_s_t),
    vert_dim="Depth_c",
    pin_cast=(i0, j0),
    pin_p=z0,
)
z_delta = z  # save for later
print(
    f" ** The in-situ density anomaly surface (referenced to {d['ref']})"
    f" intersecting the cast indexed by {(i0,j0)} at depth {z0}m"
    f" (isovalue = {d['isoval']}kg m-3)"
    f" has root-mean-square ϵ neutrality error {d['e_RMS']} kg m-4"
)

# In[Omega surfaces]

# Initialize omega surface with a (locally referenced) potential density surface.
# Provide grid distances. By default, does as many iterations as needed to get
# the root-mean-square change in the locally referenced potential density from
# one iteration to the next to drop below 10^-7 kg m-3, or a maximum of 10
# iterations.
s, t, z, d = omega_surf(
    S,
    T,
    Z,
    grid,
    vert_dim="Depth_c",
    pin_cast=(i0, j0),
    pin_p=z0,
    eos=(eos, eos_s_t),
)
s_omega, t_omega, z_omega = s, t, z  # save for later
print(
    f" ** The omega-surface"
    f" initialized from a potential density surface (referenced to {z0}m)"
    f" intersecting the cast indexed by {(i0,j0)} at depth {z0}m"
    f" has root-mean-square ϵ neutrality error {d['e_RMS'][-1]} kg m-4"
)

# Initialize omega surface with a (locally referenced) in-situ density anomaly surface.
# Use PCHIP interpolation rather than the default, linear interpolation.
# Remove the mixed layer, calculated internally according to the given parameters --
#   see `mixed_layer` for details on these parameters.
# Specify a higher max number of iterations, and quit iterations when the
# depth change from one iteration to the next has a root-mean-square value less
# than 1e-6 m.
s, t, z, d = omega_surf(
    S,
    T,
    Z,
    grid,
    ref=(None, None),
    vert_dim="Depth_c",
    pin_cast=(i0, j0),
    pin_p=z0,
    eos=(eos, eos_s_t),
    interp="pchip",
    p_ml={"bottle_index": 1, "ref_p": 0.0},
    ITER_MAX=20,
    TOL_P_CHANGE_RMS=1e-6,
    TOL_LRPD_MAV=0.0,  # deactivate this exit tolerance
    TOL_P_SOLVER=1e-7,
)
z_omega_converged = z  # save for next
print(
    f" ** The omega-surface"
    f" initialized from an in-situ density anomaly surface (referenced locally to cast {(i0,j0)} at {z0}m)"
    f" intersecting the cast indexed by {(i0,j0)} at depth {z0}m"
    f" has root-mean-square ϵ neutrality error {d['e_RMS'][-1]} kg m-4"
)

# Initialize omega surface with a pre-computed surface.
# In this case, let's continue from the above omega surface, but change the
# pinning location. Since the above omega surface is very nearly converged,
# this omega surface should basically match the above one. Indeed, it will
# do one iteration and find it is below the RMS depth change tolerance, and exit.
i1, j1 = 315, 110  # North Atlantic
s, t, z, d = omega_surf(
    S,
    T,
    Z,
    grid,
    vert_dim="Depth_c",
    pin_cast=(i1, j1),
    p_init=z_omega_converged,
    eos=(eos, eos_s_t),
    interp="pchip",
    ITER_MAX=10,
    ITER_START_WETTING=99,  # greater than ITER_MAX, so no wetting
    TOL_P_CHANGE_RMS=1e-6,
    TOL_LRPD_MAV=0.0,  # deactivate this exit tolerance
    TOL_P_SOLVER=1e-7,
)
print(
    f" ** The omega-surface"
    f" initialized from an omega surface and pinned to the cast {(i1,j1)} at {z0}m)"
    f" has root-mean-square ϵ neutrality error {d['e_RMS'][-1]} kg m-4"
)


# In[Begin showing more advanced features]

import numpy as np
import numba as nb
from neutralocean.mixed_layer import mixed_layer
from neutralocean.ntp import ntp_epsilon_errors, ntp_epsilon_errors_norms
from neutralocean.label import veronis_density
from neutralocean.lib import _process_casts
from neutralocean.ppinterp import make_pp
from neutralocean.eos import make_eos_p, vectorize_eos
from neutralocean.traj import ntp_bottle_to_cast, neutral_trajectory

# In[Show vertical interpolation]

# Build interpolation function using same interpolation kernel ("linear") as
# omega surface we saved above.
# Allow this to operate across all water columns with one call (kind="u").
# Evaluate the interpolant rather than build its coefficients (out="interp").
# Allow that there could be NaNs in the data (nans=True).
# There will be two numpy arrays of dependent data (S, T), sharing the same
# independent data Z (num_dep_vars=2).
interp_two = make_pp(
    "linear", kind="u", out="interp", nans=True, num_dep_vars=2
)

# Apply interpolation function to interpolate salinity and temperature onto the
# depth of the surface.  This requires working with numpy arrays, not xarrays.
# Evaluate the interpolant, not any of its derivatives (d=0).
s_, t_ = interp_two(z_omega.values, Z, S.values, T.values, 0)


# Check that the results of the above interpolation match (to machine precision)
# the results returned from omega_surf above.
s_check = np.allclose(s_omega.values, s_, atol=1e-15, equal_nan=True)
t_check = np.allclose(t_omega.values, t_, atol=1e-15, equal_nan=True)
if not (s_check and t_check):
    print(
        "Something's wrong; should be able to reconstruct salinity and "
        "temperature on surface by interpolating to surface's depth"
    )


# In[Veronis Density, used to label an approx neutral surface]
S_ref_cast = S.values[i0, j0]
T_ref_cast = T.values[i0, j0]
rho_v = veronis_density(S_ref_cast, T_ref_cast, Z, z0, eos="jmd95")
print(
    f"A surface through the cast indexed by {(i0,j0)} at depth {z0}m"
    f" has Veronis density {rho_v} kg m-3"
)

# In[Neutral Trajectory]

# neutral trajectory depth along meridional section
znt = np.full(nj, np.nan)

# Build a neutral trajectory going northwards from cast at (i0, j0), starting
# at depth of z0. Put result into `znt`
Snt = S[i0, j0:nj, :]
Tnt = T[i0, j0:nj, :]
s_, t_, z_ = neutral_trajectory(Snt, Tnt, Z, z0, eos=eos)
znt[j0:nj] = z_

# As above, but go southwards.
Snt = S[i0, j0:0:-1, :]
Tnt = T[i0, j0:0:-1, :]
s_, t_, z_ = neutral_trajectory(Snt, Tnt, Z, z0, eos=eos)
znt[j0:0:-1] = z_

# Consider three other trajectories along this meridional section, following ANSs
zom = z_omega[i0, :].values  # omega
zpd = z_sigma[i0, :].values  # potential density
zda = z_delta[i0, :].values  # density anomaly

# Find index to furthest south cast that has all four trajectories
j_south = np.argmax(
    np.all(
        np.isfinite(np.array((znt, zom, zpd, zda))),
        axis=0,
    )
)
print(
    f"From (i,j)={(i0,j0)} at {z0}m moving south to j={j_south}, ...\n"
    f"Neutral Trajectory              ends at {znt[j_south]:8.4f}m depth,\n"
    f"Omega Surface                   ends at {zom[j_south]:8.4f}m depth,\n"
    f"Potential Density Surface       ends at {zpd[j_south]:8.4f}m depth,\n"
    f"In-situ density anomaly surface ends at {zda[j_south]:8.4f}m depth."
)

print(
    "Uncomment below to plot neutral trajectory and ANSs along meridional section"
)
# import matplotlib.pyplot as plt
# lat = g["YCvec"]  # latitudes
# fig, ax = plt.subplots()
# ax.plot(lat, -zom, label="omega")
# ax.plot(lat, -zpd, label="potential density")
# ax.plot(lat, -zda, label="in-situ density anomaly")
# ax.plot(lat, -znt, "--k", label="neutral trajectory")
# ax.scatter(lat[j0], -z0)  # reference latitude and depth.
# ax.legend()

# In[Remove mixed layer from an omega surface]

# Pre-compute depth of the mixed layer
z_ml = mixed_layer(S, T, Z, eos)

# The correct way to remove the mixed layer from the omega surface algorithm
# is to pass the p_ml parameter, e.g. pass `p_ml=z_ml` in the call below.
# However, here we'll use a more drastic approach: setting our 3D hydrographic
# data to NaN above the mixed layer.
# This serves as a test of the new interpolation methods that allow for NaN data
# in the top of the water column followed by valid (non-NaN) data in the middle
# where the interpolation is done, followed by more NaN data at the bottom.
# This is to allow for models or data with ice shelves.
# Note, this drastic NaN'ing of the input (S,T) data is not expected to give
# exactly the same results as by passing the p_ml parameter, since
# (a) with p_ml, the mixed layer removal is not applied until after the first
#     iteration, and
# (b) with PCHIPs, the interpolant just below the mixed layer is affected by
#     a few of the lower bottles in the mixed layer.
# (c) with p_ml, the surface is set to NaN only if it goes above p_ml which
#     is any real number, whereas with the data-NaN'ing approach, discrete
#     bottles are set to NaN so the surface will only be valid if it is below
#     the first valid bottle below what was NaN'ed out.


@nb.njit
def mixedlayer2nan(S, T, Z, z_ml):
    for n in np.ndindex(z_ml.shape):
        z = z_ml[n]
        for k in range(nk):
            if Z[k] < z:
                S[(*n, k)] = np.nan
                T[(*n, k)] = np.nan
            else:
                break


mixedlayer2nan(S.values, T.values, Z, z_ml)


s, t, z, d = omega_surf(
    S,
    T,
    Z,
    grid,
    ref=(None, None),
    vert_dim="Depth_c",
    pin_cast=(i0, j0),
    pin_p=z0,
    eos=(eos, eos_s_t),
    interp="pchip",
    ITER_MAX=10,
    ITER_START_WETTING=1,
    TOL_P_SOLVER=1e-5,
)

# potential and anomaly surfaces don't take `p_ml` as an argument, since they
# aren't iterative algorithms, so we can just remove it manually, e.g.
# z[z < z_ml] = np.nan

# In[Neutrality errors on a surface]
e_RMS, e_MAV = ntp_epsilon_errors_norms(s, t, z_omega, grid, eos_s_t)
print(f"RMS of ϵ is {e_RMS : 4e} [kg m-4])")

# Calculate ϵ neutrality errors on all pairs of adjacent water columns
e = ntp_epsilon_errors(s, t, z, grid, eos_s_t)

# Convert ϵ above into two 2D maps, one for zonal ϵ errors and one for meridional ϵ errors
ex, ey = edgedata_to_maps(e, (ni, nj), g["wrap"])
# These can then be mapped...

# In[Neutral Tangent Plane bottle to cast]

sB, tB, zB = 35.0, 16.0, 500.0  # Thermodynamic properties of a given Bottle
S1 = S.values[180, 80, :]
T1 = T.values[180, 80, :]
s1, t1, z1 = ntp_bottle_to_cast(sB, tB, zB, S1, T1, Z)

# In[Work with Numpy arrays instead of xarrays]

# Convert S and T from xarray to numpy ndarrays, and make vertical dimension
# contiguous in memory.  If not done here in advance, this will be done each
# time potential_surf, anomaly_surf, or omega_surf is called.  Hence, this is
# the recommended method if many approx neutral surfaces will be calculated.
Snp, Tnp, Znp = _process_casts(S, T, Z, "Depth_c")

s, t, z, d = anomaly_surf(
    Snp,
    Tnp,
    Znp,
    grid=grid,
    eos=(eos, eos_s_t),
    vert_dim=-1,
    ref=(s0, t0),
    isoval=0.0,
)


# In[Calculate a large-scale potential vorticity on the above specvol anomaly surface]

# Create function for partial deriv of equation of state with respect to depth z
eos_z = make_eos_p("jmd95", g["grav"], g["rho_c"])  # for scalar inputs
eos_z = vectorize_eos(eos_z)  # for nd inputs

# Build linear interpolation function, operating over a whole dataset (kind="u"),
# evaluating the interpolant on the fly (out="interp"), for two dependent
# variables (num_dep_vars=2).
interp_two = make_pp("linear", kind="u", out="interp", num_dep_vars=2)

# Earth sidereal day period [s]
Earth_day = 86164

# Coriolis param [s-1] on tracer grid
f = 2 * (2 * np.pi / Earth_day) * np.sin(g["YCvec"] * (np.pi / 180))

# evaluate first derivative (d=1) of S(z) and T(z) at depth z.  That is, eval
# ∂S/∂Z and ∂T/∂Z, on the surface.
sz, tz = interp_two(z, Z, S.values, T.values, 1)
rs, rt = eos_s_t(s, t, z)  # ∂ρ/∂S and ∂ρ/∂T, on the surface

# ∂δ/∂z on the surface, where δ is the in-situ density anomaly
δz = rs * sz + rt * tz + (eos_z(s, t, z) - eos_z(s0, t0, z))

# large-scale potential vorticity [m-1 s-1] defined via in-situ density anomaly, on the surface
q = f * δz
